# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import GLib
from gi.repository import Gtk


class EmojiChooser(Gtk.EmojiChooser):
    def __init__(self) -> None:
        Gtk.EmojiChooser.__init__(self)

        self._emoji_picked_func: Any = None

        self.connect("closed", self._on_closed)

    def set_emoji_picked_func(self, func: Any) -> None:
        self._emoji_picked_func = func
        self.connect("emoji-picked", self._emoji_picked_func)

    def _on_closed(self, _popover: Gtk.EmojiChooser) -> None:
        def _cleanup() -> None:
            self.disconnect_by_func(self._emoji_picked_func)
            self._emoji_picked_func = None

        # We don't want to assume the 'emoji-picked' signal
        # is raised before 'closed'
        GLib.idle_add(_cleanup)



# Copyright (C) 2009 Thibaut GIRKA <thib AT sitedethib.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

try:
    from gi.repository import Gst
except Exception:
    pass

from gajim.common.i18n import _

log = logging.getLogger("gajim.c.util.av")


class DeviceManager:
    def __init__(self) -> None:
        self.devices: dict[str, str] = {}

    def detect(self) -> None:
        self.devices = {}

    def get_devices(self) -> dict[str, str]:
        if not self.devices:
            self.detect()
        return self.devices

    def detect_element(self, name: str, text: str, pipe: str = "%s") -> None:
        if Gst.ElementFactory.find(name):
            element: Gst.Element | None = Gst.ElementFactory.make(
                name, f"{name}presencetest"
            )
            if element is None:
                log.warning("could not create %spresencetest", name)
                return

            if hasattr(element.props, "device"):
                element.set_state(Gst.State.READY)
                devices = element.get_properties("device")
                if devices:
                    self.devices[text % _("Default device")] = pipe % name
                    for device in devices:
                        if device is None:
                            continue
                        element.set_state(Gst.State.NULL)
                        element.set_property("device", device)
                        element.set_state(Gst.State.READY)
                        device_name = element.get_property("device-name")
                        self.devices[text % device_name] = (
                            pipe % f"{name} device={device}"
                        )
                element.set_state(Gst.State.NULL)
            else:
                self.devices[text] = pipe % name
        else:
            log.info("element %s not found", name)


class AudioInputManager(DeviceManager):
    def detect(self) -> None:
        self.devices: dict[str, str] = {}
        # Test src
        self.detect_element(
            "audiotestsrc", _("Audio test"), "%s is-live=true name=gajim_vol"
        )
        # Auto src
        self.detect_element(
            "autoaudiosrc", _("Autodetect"), "%s ! volume name=gajim_vol"
        )
        # Alsa src
        self.detect_element("alsasrc", _("ALSA: %s"), "%s ! volume name=gajim_vol")
        # Pulseaudio src
        self.detect_element("pulsesrc", _("Pulse: %s"), "%s ! volume name=gajim_vol")


class AudioOutputManager(DeviceManager):
    def detect(self) -> None:
        self.devices: dict[str, str] = {}
        # Fake sink
        self.detect_element("fakesink", _("Fake audio output"))
        # Auto sink
        self.detect_element("autoaudiosink", _("Autodetect"))
        # Alsa sink
        self.detect_element("alsasink", _("ALSA: %s"), "%s sync=false")
        # Pulseaudio sink
        self.detect_element("pulsesink", _("Pulse: %s"), "%s sync=true")


class VideoInputManager(DeviceManager):
    def detect(self) -> None:
        self.devices: dict[str, str] = {}
        # Test src
        self.detect_element(
            "videotestsrc",
            _("Video test"),
            "%s is-live=true ! video/x-raw,framerate=10/1 ! videoconvert",
        )
        # Auto src
        self.detect_element("autovideosrc", _("Autodetect"))
        # Best source on Linux, for both camera and screen sharing
        self.detect_element("pipewiresrc", _("Pipewire"))
        # Camera source on Linux
        self.detect_element("v4l2src", _("V4L2: %s"))
        # X11 screen sharing on Linux
        self.detect_element("ximagesrc", _("X11"))
        # Recommended source on Windows
        self.detect_element("ksvideosrc", _("Windows"))
        # Recommended source on OS X
        self.detect_element("avfvideosrc", _("macOS"))



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload

import itertools
import logging

from gi.repository import Gdk
from gi.repository import Gtk
from nbxmpp.modules import dataforms
from nbxmpp.simplexml import Node

from gajim.common import app
from gajim.common import ged
from gajim.common.events import SearchFormReceivedEvent
from gajim.common.events import SearchResultReceivedEvent
from gajim.common.i18n import _

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.assistant import AssistantProgressPage
from gajim.gtk.dataform import DataFormWidget
from gajim.gtk.menus import get_component_search_menu
from gajim.gtk.util.misc import ensure_not_destroyed
from gajim.gtk.widgets import GajimPopover

log = logging.getLogger("gajim.gtk.search")


class ComponentSearch(Assistant):
    def __init__(
        self, account: str, jid: str, transient_for: Gtk.Window | None = None
    ) -> None:
        Assistant.__init__(self, transient_for=transient_for, width=700, height=500)

        self._client = app.get_client(account)
        self.account = account
        self._jid = jid
        self._destroyed = False

        self.add_button("close", _("Close"))
        self.add_button("search", _("Search"), "suggested-action")
        self.add_button("new-search", _("New Search"))

        self.add_pages(
            {
                "prepare": RequestForm(),
                "form": SearchForm(),
                "result": Result(),
                "error": Error(),
            }
        )

        progress = self.add_default_page("progress")
        progress.set_title(_("Searching"))
        progress.set_text(_("Searching…"))

        self._connect(self, "button-clicked", self._on_button_clicked)

        self.register_events(
            [
                ("search-form-received", ged.GUI1, self._search_form_received),
                ("search-result-received", ged.GUI1, self._search_result_received),
            ]
        )

        self._client.get_module("Search").request_search_fields(self._jid)

    @overload
    def get_page(self, name: Literal["prepare"]) -> RequestForm: ...

    @overload
    def get_page(self, name: Literal["form"]) -> SearchForm: ...

    @overload
    def get_page(self, name: Literal["result"]) -> Result: ...

    @overload
    def get_page(self, name: Literal["error"]) -> Error: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    def _on_button_clicked(self, _assistant: Assistant, button_name: str) -> None:
        if button_name == "search":
            self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)
            form = self.get_page("form").get_submit_form()
            self._client.get_module("Search").send_search_form(self._jid, form, True)
            return

        if button_name == "new-search":
            self.show_page("form", Gtk.StackTransitionType.SLIDE_RIGHT)
            return

        if button_name == "close":
            self.close()

    @ensure_not_destroyed
    def _search_form_received(self, event: SearchFormReceivedEvent) -> None:
        if not event.is_dataform:
            self.get_page("error").set_text(_("Error while retrieving search form."))
            self.show_page("error")
            return

        self.get_page("form").process_search_form(event.data)
        self.show_page("form")

    @ensure_not_destroyed
    def _search_result_received(self, event: SearchResultReceivedEvent) -> None:
        if event.data is None:
            self.get_page("error").set_text(_("Error while receiving search results."))
            self.show_page("error")
            return

        self.get_page("result").process_result(event.data)
        self.show_page("result")

    def _cleanup(self) -> None:
        Assistant._cleanup(self)
        self._destroyed = True


class RequestForm(AssistantProgressPage):
    def __init__(self):
        AssistantProgressPage.__init__(self)
        self.set_title(_("Request Search Form"))
        self.set_text(_("Requesting search form from server"))

    def get_visible_buttons(self) -> list[str]:
        return ["close"]


class SearchForm(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Search")

        self.complete = False

        self._dataform_widget = None

    @property
    def search_form(self) -> dataforms.SimpleDataForm:
        assert self._dataform_widget is not None
        return self._dataform_widget.get_submit_form()

    def clear(self) -> None:
        self._show_form(None)

    def process_search_form(self, form: Node) -> None:
        self._show_form(form)

    def _show_form(self, form: Node | None) -> None:
        if self._dataform_widget is not None:
            self.remove(self._dataform_widget)
        if form is None:
            return

        options = {"form-width": 350, "entry-activates-default": True}

        form = dataforms.extend_form(node=form)
        self._dataform_widget = DataFormWidget(form, options=options)
        self._dataform_widget.set_propagate_natural_height(True)
        self._connect(self._dataform_widget, "is-valid", self._on_is_valid)
        self._dataform_widget.validate()
        self.append(self._dataform_widget)

    def _on_is_valid(self, _widget: DataFormWidget, is_valid: bool) -> None:
        self.complete = is_valid
        self.update_page_complete()

    def get_submit_form(self) -> dataforms.SimpleDataForm:
        assert self._dataform_widget is not None
        return self._dataform_widget.get_submit_form()

    def get_visible_buttons(self) -> list[str]:
        return ["close", "search"]

    def get_default_button(self) -> str:
        return "search"


class Result(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Search Result")

        self._jid_col: int | None = None

        self._label = Gtk.Label(label=_("No results found"))
        self._label.add_css_class("title-3")
        self._label.set_visible(False)
        self._label.set_halign(Gtk.Align.CENTER)

        self._scrolled = Gtk.ScrolledWindow(
            propagate_natural_height=True,
            visible=False,
            has_frame=True,
        )

        self.append(self._label)
        self.append(self._scrolled)

        self._treeview: Gtk.TreeView | None = None

    def process_result(self, form: Node | None) -> None:
        if self._treeview is not None:
            self._scrolled.set_child(None)
            self._treeview = None
            self._label.set_visible(False)
            self._scrolled.set_visible(False)

        if not form:
            self._label.set_visible(True)
            return

        form = dataforms.extend_form(node=form)
        assert isinstance(form, dataforms.MultipleDataForm)

        fieldtypes: list[type[bool] | type[str]] = []
        fieldvars: list[Any] = []
        index = 0
        for field in form.reported.iter_fields():
            if field.type_ == "boolean":
                fieldtypes.append(bool)
            elif field.type_ in ("jid-single", "text-single"):
                fieldtypes.append(str)
                if field.type_ == "jid-single" or field.var == "jid":
                    # Enable Start Chat context menu entry
                    self._jid_col = index
            else:
                log.warning("Not supported field received: %s", field.type_)
                continue
            fieldvars.append(field.var)
            index += 1

        liststore = Gtk.ListStore(*fieldtypes)

        for item in form.iter_records():
            iter_ = liststore.append()
            assert isinstance(item, dataforms.DataRecord)
            for field in item.iter_fields():
                if field.var in fieldvars:
                    liststore.set_value(iter_, fieldvars.index(field.var), field.value)

        self._treeview = Gtk.TreeView()
        self._treeview.set_hexpand(True)
        self._treeview.set_vexpand(True)
        self._treeview.add_css_class("gajim-treeview")

        gesture_secondary = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_secondary, "pressed", self._on_button_press)
        self._treeview.add_controller(gesture_secondary)

        for field, counter in zip(
            form.reported.iter_fields(), itertools.count(), strict=False
        ):
            self._treeview.append_column(
                Gtk.TreeViewColumn(field.label, Gtk.CellRendererText(), text=counter)
            )

        self._treeview.set_model(liststore)
        self._treeview.set_visible(True)

        self._scrolled.set_visible(True)

        box = Gtk.Box()
        box.append(self._treeview)

        self._popover_menu = GajimPopover(None)
        box.append(self._popover_menu)
        self._scrolled.set_child(box)

    def _on_button_press(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        assert self._treeview is not None
        path = self._treeview.get_path_at_pos(int(x), int(y))
        if path is None:
            return

        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)

        path, _column, _x, _y = path
        store = self._treeview.get_model()
        assert store is not None
        assert path is not None
        iter_ = store.get_iter(path)
        column_values = [str(item) for item in store[iter_]]
        text = " ".join(column_values)

        jid = None
        if self._jid_col is not None:
            jid = column_values[self._jid_col]
        menu = get_component_search_menu(jid, text)
        self._popover_menu.set_menu_model(menu)
        self._popover_menu.set_pointing_to_coord(x, y)
        self._popover_menu.popup()

    def get_visible_buttons(self) -> list[str]:
        return ["close", "new-search"]

    def get_default_button(self) -> str:
        return "close"


class Error(AssistantErrorPage):
    def __init__(self) -> None:
        AssistantErrorPage.__init__(self)
        self.set_title(_("Error"))
        self.set_heading(_("An error occurred"))

    def get_visible_buttons(self) -> list[str]:
        return ["close"]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from typing import TYPE_CHECKING

import math
from pathlib import Path

import gi

try:
    gi.require_version("Gst", "1.0")
    from gi.repository import Gst
except Exception:
    if TYPE_CHECKING:
        from gi.repository import Gst

Gst.init(None)


def extract_audio_properties(
    input_path: Path,
) -> tuple[list[tuple[float, float]], float] | None:
    playbin = Gst.ElementFactory.make("playbin")
    audio_sink = Gst.Bin.new("audiosink")
    audioconvert = Gst.ElementFactory.make("audioconvert")
    level = Gst.ElementFactory.make("level")
    fakesink = Gst.ElementFactory.make("fakesink")

    pipeline_elements = [
        playbin,
        audio_sink,
        audioconvert,
        level,
        fakesink,
    ]

    if any(element is None for element in pipeline_elements):
        raise Exception(
            f"\n{__name__}: Could not set up pipeline for audio preview analyzer"
        )

    assert audioconvert is not None
    assert level is not None
    assert fakesink is not None

    audio_sink.add(audioconvert)
    audio_sink.add(level)
    audio_sink.add(fakesink)

    audioconvert.link(level)
    level.link(fakesink)

    fakesink.set_property("sync", False)

    samples: list[tuple[float, float]] = []
    duration: float = 0.0

    sink_pad = audioconvert.get_static_pad("sink")
    assert sink_pad is not None
    ghost_pad = Gst.GhostPad.new("sink", sink_pad)
    assert ghost_pad is not None
    audio_sink.add_pad(ghost_pad)

    assert playbin is not None
    playbin.set_property("audio-sink", audio_sink)
    playbin.set_property("uri", input_path.as_uri())
    playbin.no_more_pads()

    state_return = playbin.set_state(Gst.State.PLAYING)
    if state_return == Gst.StateChangeReturn.FAILURE:
        raise Exception(f"\n{__name__}: Failed to set GST playbin to PLAYING")

    bus = playbin.get_bus()
    if not bus:
        raise Exception(f"\n{__name__}: Could not get GST bus")

    while True:
        msg = bus.timed_pop_filtered(
            Gst.CLOCK_TIME_NONE,
            Gst.MessageType.EOS
            | Gst.MessageType.ERROR
            | Gst.MessageType.STATE_CHANGED
            | Gst.MessageType.DURATION_CHANGED
            | Gst.MessageType.ELEMENT,
        )

        if msg is None:
            break

        if msg.type == Gst.MessageType.ERROR:
            err, debug = msg.parse_error()
            raise Exception(f"\n{__name__}: Error: {err}, debug: {debug}")

        if msg.type == Gst.MessageType.EOS:
            break

        if msg.type in (
            Gst.MessageType.STATE_CHANGED,
            Gst.MessageType.DURATION_CHANGED,
        ):
            _success, duration = playbin.query_duration(Gst.Format.TIME)

        if msg.type == Gst.MessageType.ELEMENT:
            structure = msg.get_structure()
            if (
                structure
                and structure.get_name() == "level"
                and structure.has_field("rms")
            ):
                rms_values = structure.get_value("rms")
                if rms_values:
                    num_channels = min(2, len(rms_values))
                    if num_channels == 0:
                        samples.append((0.0, 0.0))
                    elif num_channels == 1:
                        lin_val = math.pow(10, rms_values[0] / 10 / 2)
                        samples.append((lin_val, lin_val))
                    else:
                        lin_val1 = math.pow(10, rms_values[0] / 10 / 2)
                        lin_val2 = math.pow(10, rms_values[1] / 10 / 2)
                        samples.append((lin_val1, lin_val2))

    playbin.set_state(Gst.State.NULL)

    return samples, duration



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

import argparse
import logging
import sys

from gi.repository import Gio
from gi.repository import GLib

LOG_FORMAT = "%(asctime)s %(levelname)s %(message)s"
logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
log = logging.getLogger()


OBJ_PATH = "/org/gajim/dbus/RemoteObject"
INTERFACE = "org.gajim.dbus.RemoteInterface"
SERVICE = "org.gajim.Gajim"

SIGNATURES = {
    "list_contacts": "(s)",
    "list_accounts": "()",
    "change_status": "(sss)",
    "send_chat_message": "(sss)",
    "send_groupchat_message": "(sss)",
    "account_info": "(s)",
    "get_status": "(s)",
    "get_status_message": "(s)",
    "get_unread_msgs_number": "()",
}


def call_method(args: argparse.Namespace) -> Any:
    arg_dict = vars(args)
    app_id = arg_dict.pop("app_id")
    command = arg_dict.pop("command")

    proxy = Gio.DBusProxy.new_for_bus_sync(
        Gio.BusType.SESSION,
        Gio.DBusProxyFlags.NONE,
        None,
        app_id,
        OBJ_PATH,
        INTERFACE,
        None,
    )

    arguments = tuple(arg_dict.values())
    signature = SIGNATURES[command]
    method = getattr(proxy, command)
    return method(signature, *arguments)


def create_arg_parser() -> argparse.ArgumentParser:
    account_help = "The account the command is executed for"

    parser = argparse.ArgumentParser(prog="gajim-remote")
    parser.add_argument("--app-id", default=SERVICE)

    subparsers = parser.add_subparsers(
        required=True, metavar="commands", dest="command"
    )

    subparser = subparsers.add_parser("list_contacts", help="Get all roster contacts")
    subparser.add_argument("account", type=str)

    subparser = subparsers.add_parser("list_accounts", help="Get the list of accounts")

    subparser = subparsers.add_parser("change_status", help="Change the status")
    subparser.add_argument("status", choices=["offline", "online", "away", "xa", "dnd"])
    subparser.add_argument("message", type=str)
    subparser.add_argument("account", type=str)

    subparser = subparsers.add_parser(
        "send_chat_message", help="Send a chat message to a contact"
    )
    subparser.add_argument("address", type=str, help="The XMPP address of the contact")
    subparser.add_argument("message", type=str, help="The message to be sent")
    subparser.add_argument("account", type=str, help=account_help)

    subparser = subparsers.add_parser(
        "send_groupchat_message", help="Send a chat message to a group chat"
    )
    subparser.add_argument(
        "address", type=str, help="The XMPP address of the group chat"
    )
    subparser.add_argument("message", type=str, help="The message to be sent")
    subparser.add_argument("account", type=str, help=account_help)

    subparser = subparsers.add_parser("account_info", help="Get account details")
    subparser.add_argument("account", type=str, help=account_help)

    subparser = subparsers.add_parser("get_status", help="Get the current status")
    subparser.add_argument("account", type=str, help=account_help)

    subparser = subparsers.add_parser(
        "get_status_message", help="Get the current status message"
    )
    subparser.add_argument("account", type=str, help=account_help)

    subparser = subparsers.add_parser(
        "get_unread_msgs_number", help="Get the unread message count"
    )

    return parser


def run() -> None:
    args = create_arg_parser().parse_args()
    try:
        result = call_method(args)
    except GLib.Error as error:
        quark = GLib.quark_try_string("g-dbus-error-quark")
        if error.matches(quark, Gio.DBusError.SERVICE_UNKNOWN):
            log.error(
                "Service not found. Check if Gajim is running. "
                "If Gajim is running under a custom profile use "
                "--app-id=org.gajim.Gajim.myprofilename"
            )

        else:
            log.exception("Failed to execute method")

        sys.exit(1)

    except Exception:
        log.exception("Failed to execute method")
        sys.exit(1)

    print(result)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0199: XMPP Ping

from __future__ import annotations

from typing import Any

import time
from collections.abc import Generator

from nbxmpp.errors import is_error
from nbxmpp.structs import CommonResult

from gajim.common import app
from gajim.common import types
from gajim.common.events import PingError
from gajim.common.events import PingReply
from gajim.common.events import PingSent
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import as_task


class Ping(BaseModule):

    _nbxmpp_extends = 'Ping'
    _nbxmpp_methods = [
        'ping',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = []

    @as_task
    def send_ping(self,
                  contact: types.ContactT
                  ) -> Generator[Any, Any]:
        _task = yield

        if not app.account_is_available(self._account):
            return

        jid = contact.jid

        self._log.info('Send ping to %s', str(jid))

        app.ged.raise_event(PingSent(account=self._account, contact=contact))

        ping_time = time.time()

        response = yield self.ping(jid, timeout=10)
        if is_error(response):
            app.ged.raise_event(PingError(
                account=self._account,
                contact=contact,
                error=str(response)))
            return

        assert isinstance(response, CommonResult)

        diff = round(time.time() - ping_time, 2)
        self._log.info('Received pong from %s after %s seconds',
                       response.jid, diff)

        app.ged.raise_event(PingReply(account=self._account,
                                      contact=contact,
                                      seconds=diff))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


# Handles  Jingle File Transfer (XEP 0234)


from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import logging
import threading
import uuid
from enum import IntEnum
from enum import unique

import nbxmpp
from nbxmpp import JID
from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common import helpers

# from gajim.common.events import FileRequestReceivedEvent
from gajim.common.file_props import FileProp
from gajim.common.file_props import FilesProp
from gajim.common.jingle_content import contents
from gajim.common.jingle_content import JingleContent
from gajim.common.jingle_ftstates import StateCandReceived
from gajim.common.jingle_ftstates import StateCandSent
from gajim.common.jingle_ftstates import StateCandSentAndRecv
from gajim.common.jingle_ftstates import StateInitialized
from gajim.common.jingle_ftstates import StateTransfering
from gajim.common.jingle_ftstates import StateTransportReplace
from gajim.common.jingle_transport import JingleTransportSocks5
from gajim.common.jingle_transport import TransportType
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.util.datetime import utc_now

if TYPE_CHECKING:
    from gajim.common.jingle_session import JingleSession
    from gajim.common.jingle_transport import JingleTransport

log = logging.getLogger('gajim.c.jingle_ft')


@unique
class State(IntEnum):
    NOT_STARTED = 0
    INITIALIZED = 1
    # We send the candidates and we are waiting for a reply
    CAND_SENT = 2
    # We received the candidates and we are waiting to reply
    CAND_RECEIVED = 3
    # We have sent and received the candidates
    # This also includes any candidate-error received or sent
    CAND_SENT_AND_RECEIVED = 4
    TRANSPORT_REPLACE = 5
    # We are transferring the file
    TRANSFERRING = 6


class JingleFileTransfer(JingleContent):
    def __init__(self,
                 session: JingleSession,
                 file_props: FileProp,
                 transport: JingleTransport | None = None,
                 senders: str | None = None
                 ) -> None:

        JingleContent.__init__(self, session, transport, senders)
        log.info('transport value: %s', transport)
        # events we might be interested in
        self.callbacks['session-initiate'] += [self.__on_session_initiate]
        self.callbacks['session-initiate-sent'] += [
            self.__on_session_initiate_sent]
        self.callbacks['content-add'] += [self.__on_session_initiate]
        self.callbacks['session-accept'] += [self.__on_session_accept]
        self.callbacks['session-terminate'] += [self.__on_session_terminate]
        self.callbacks['session-info'] += [self.__on_session_info]
        self.callbacks['transport-accept'] += [self.__on_transport_accept]
        self.callbacks['transport-replace'] += [self.__on_transport_replace]
        self.callbacks['session-accept-sent'] += [self.__transport_setup]
        # fallback transport method
        self.callbacks['transport-reject'] += [self.__on_transport_reject]
        self.callbacks['transport-info'] += [self.__on_transport_info]
        self.callbacks['iq-result'] += [self.__on_iq_result]

        self.file_props = file_props
        self.weinitiate = self.session.weinitiate
        self.werequest = self.session.werequest

        if transport is None:
            self.transport = JingleTransportSocks5()
        self.transport.set_connection(session.connection)
        self.transport.set_file_props(self.file_props)
        self.transport.set_our_jid(session.ourjid)
        log.info('ourjid: %s', session.ourjid)

        if self.session.werequest:
            self.file_props.sender = self.session.peerjid
            self.file_props.receiver = self.session.ourjid
        else:
            self.file_props.sender = self.session.ourjid
            self.file_props.receiver = self.session.peerjid
        self.file_props.session_type = 'jingle'
        self.file_props.sid = session.sid
        self.file_props.transferred_size = []
        self.file_props.transport_sid = self.transport.sid

        self.session = session
        self.media = 'file'
        self.nominated_cand = {}
        self.state = State.NOT_STARTED
        self.states = {
            State.INITIALIZED: StateInitialized(self),
            State.CAND_SENT: StateCandSent(self),
            State.CAND_RECEIVED: StateCandReceived(self),
            State.TRANSFERRING: StateTransfering(self),
            State.TRANSPORT_REPLACE: StateTransportReplace(self),
            State.CAND_SENT_AND_RECEIVED: StateCandSentAndRecv(self)
        }

    def __state_changed(self,
                        nextstate: State,
                        args: dict[str, Any] | None = None
                        ) -> None:
        # Executes the next state action and sets the next state
        current_state = self.state
        st = self.states[nextstate]
        st.action(args)
        # state can have been changed during the action. Don't go back.
        if self.state == current_state:
            self.state = nextstate

    def __on_session_initiate(self,
                              stanza: nbxmpp.Node,
                              content: nbxmpp.Node,
                              error: nbxmpp.Node | None,
                              action: str
                              ) -> None:
        log.debug('Jingle FT request received')
        self._raise_event(stanza, content)

        account = self.session.connection.name
        jid = self.session.connection.get_module('Bytestream')._ft_get_from(
            stanza)
        jid = JID.from_string(jid)
        sid = stanza.getTag('jingle').getAttr('sid')
        assert sid is not None

        ft_data = mod.FileTransfer(
            state=1,
            source=[mod.JingleFT(type='jingleft', sid=sid)],
        )

        message_data = mod.Message(
            account_=account,
            remote_jid_=jid.new_as_bare(),
            resource=jid.resource,
            type=MessageType.CHAT,
            direction=ChatDirection.INCOMING,
            timestamp=utc_now(),
            state=MessageState.ACKNOWLEDGED,
            id=str(uuid.uuid4()),
            filetransfers=[ft_data],
        )

        app.storage.archive.insert_object(message_data)

        if self.session.request:
            # accept the request
            self.session.approve_content(self.media, self.name)
            self.session.accept_session()

    def _raise_event(self,
                     stanza: nbxmpp.Node,
                     content: nbxmpp.Node
                     ) -> None:
        con = self.session.connection
        id_ = stanza.getID()
        fjid = con.get_module('Bytestream')._ft_get_from(stanza)
        account = con.name
        jid = app.get_jid_without_resource(fjid)
        jid = JID.from_string(jid)
        if not content:
            return

        if not self.transport:
            self.transport = JingleTransportSocks5()
            self.transport.set_our_jid(self.session.ourjid)
            self.transport.set_connection(con)

        sid = stanza.getTag('jingle').getAttr('sid')
        file_props = FilesProp.getNewFileProp(account, sid)
        file_props.transport_sid = self.transport.sid

        self.transport.set_file_props(file_props)
        file_props.streamhosts.extend(self.transport.remote_candidates)

        for host in file_props.streamhosts:
            host['initiator'] = self.session.initiator
            host['target'] = self.session.responder

        file_props.session_type = 'jingle'
        desc = content.getTag('description')
        if content.getAttr('creator') == 'initiator':
            file_tag = desc.getTag('file')
            file_props.sender = fjid
            file_props.receiver = str(con.get_own_jid())
        else:
            file_tag = desc.getTag('file')
            hash_ = file_tag.getTag('hash')
            hash_ = hash_.getData() if hash_ else None
            file_name = file_tag.getTag('name')
            file_name = file_name.getData() if file_name else None
            pjid = app.get_jid_without_resource(fjid)
            file_info = con.get_module('Jingle').get_file_info(
                pjid, hash_=hash_, name=file_name, account=account)
            file_props.file_name = file_info['file-name']
            file_props.sender = str(con.get_own_jid())
            file_props.receiver = fjid
            file_props.type_ = 's'

        for child in file_tag.getChildren():
            name = child.getName()
            val = child.getData()
            if val is None:
                continue
            if name == 'name':
                file_props.name = val
            if name == 'size':
                file_props.size = int(val)
            if name == 'hash':
                file_props.algo = child.getAttr('algo')
                file_props.hash_ = val
            if name == 'date':
                file_props.date = val

        file_props.request_id = id_
        file_desc_tag = file_tag.getTag('desc')
        if file_desc_tag is not None:
            file_props.desc = file_desc_tag.getData()
        file_props.transferred_size = []

        # TODO Jingle FT
        # app.ged.raise_event(FileRequestReceivedEvent(
        #     conn=con,
        #     stanza=stanza,
        #     id_=id_,
        #     fjid=fjid,
        #     account=account,
        #     jid=jid,
        #     file_props=file_props))

    def __on_session_initiate_sent(self,
                                   stanza: nbxmpp.Node,
                                   content: nbxmpp.Node,
                                   error: nbxmpp.Node | None,
                                   action: str
                                   ) -> None:
        pass

    def __send_hash(self) -> None:
        # Send hash in a session info
        checksum = nbxmpp.Node(tag='checksum',
                               payload=[
                                   nbxmpp.Node(tag='file',
                                               payload=[self._compute_hash()])
                               ])
        checksum.setNamespace(Namespace.JINGLE_FILE_TRANSFER_5)
        self.session.__session_info(checksum)
        pjid = app.get_jid_without_resource(self.session.peerjid)
        file_info = {
            'name': self.file_props.name,
            'file-name': self.file_props.file_name,
            'hash': self.file_props.hash_,
            'size': self.file_props.size,
            'date': self.file_props.date,
            'peerjid': pjid
        }
        self.session.connection.get_module('Jingle').set_file_info(file_info)

    def _compute_hash(self) -> nbxmpp.Hashes2 | None:
        # Calculates the hash and returns a xep-300 hash stanza
        if self.file_props.algo is None:
            return
        try:
            file_ = open(self.file_props.file_name, 'rb')
        except OSError:
            # can't open file
            return
        h = nbxmpp.Hashes2()
        hash_ = h.calculateHash(self.file_props.algo, file_)
        file_.close()
        # DEBUG
        # hash_ = '1294809248109223'
        if not hash_:
            # Hash algorithm not supported
            return
        self.file_props.hash_ = hash_
        h.addHash(hash_, self.file_props.algo)
        return h

    def on_cert_received(self) -> None:
        self.session.approve_session()
        self.session.approve_content('file', name=self.name)

    def __on_session_accept(self,
                            stanza: nbxmpp.Node,
                            content: nbxmpp.Node,
                            error: nbxmpp.Node | None,
                            action: str
                            ) -> None:
        log.info('__on_session_accept')
        if self.state == State.TRANSPORT_REPLACE:
            # If we are requesting we don't have the file
            if self.session.werequest:
                raise nbxmpp.NodeProcessed
            # We send the file
            self.__state_changed(State.TRANSFERRING)
            raise nbxmpp.NodeProcessed
        self.file_props.streamhosts = self.transport.remote_candidates
        # Calculate file hash in a new thread
        # if we haven't sent the hash already.
        if self.file_props.hash_ is None and self.file_props.algo and \
                not self.werequest:
            self.hash_thread = threading.Thread(target=self.__send_hash)
            self.hash_thread.start()
        for host in self.file_props.streamhosts:
            host['initiator'] = self.session.initiator
            host['target'] = self.session.responder
            host['sid'] = self.file_props.sid
        if self.transport.type_ == TransportType.SOCKS5:
            sid = self.file_props.transport_sid
            app.socks5queue.connect_to_hosts(self.session.connection.name,
                                             sid,
                                             self.on_connect,
                                             self._on_connect_error,
                                             receiving=False)
            raise nbxmpp.NodeProcessed
        self.__state_changed(State.TRANSFERRING)
        raise nbxmpp.NodeProcessed

    def __on_session_terminate(self,
                               stanza: nbxmpp.Node,
                               content: nbxmpp.Node,
                               error: nbxmpp.Node | None,
                               action: str
                               ) -> None:
        log.info('__on_session_terminate')

    def __on_session_info(self,
                          stanza: nbxmpp.Node,
                          content: nbxmpp.Node,
                          error: nbxmpp.Node | None,
                          action: str
                          ) -> None:
        pass

    def __on_transport_accept(self,
                              stanza: nbxmpp.Node,
                              content: nbxmpp.Node,
                              error: nbxmpp.Node | None,
                              action: str
                              ) -> None:
        log.info('__on_transport_accept')
        if content.getTag('transport').getNamespace() == Namespace.JINGLE_IBB:
            self.__state_changed(State.TRANSFERRING)

    def __on_transport_replace(self,
                               stanza: nbxmpp.Node,
                               content: nbxmpp.Node,
                               error: nbxmpp.Node | None,
                               action: str
                               ) -> None:
        log.info('__on_transport_replace')

    def __on_transport_reject(self,
                              stanza: nbxmpp.Node,
                              content: nbxmpp.Node,
                              error: nbxmpp.Node | None,
                              action: str
                              ) -> None:
        log.info('__on_transport_reject')

    def __on_transport_info(self,
                            stanza: nbxmpp.Node,
                            content: nbxmpp.Node,
                            error: nbxmpp.Node | None,
                            action: str
                            ) -> None:
        log.info('__on_transport_info')
        cand_error = content.getTag('transport').getTag('candidate-error')
        cand_used = content.getTag('transport').getTag('candidate-used')
        if (cand_error or cand_used) and \
                self.state >= State.CAND_SENT_AND_RECEIVED:
            raise nbxmpp.NodeProcessed
        if cand_error:
            if not app.socks5queue.listener.connections:
                app.socks5queue.listener.disconnect()
            self.nominated_cand['peer-cand'] = False
            if self.state == State.CAND_SENT:
                if not self.nominated_cand['our-cand'] and \
                   not self.nominated_cand['peer-cand']:
                    if not self.weinitiate:
                        return
                    self.__state_changed(State.TRANSPORT_REPLACE)
                else:
                    response = stanza.buildReply('result')
                    response.delChild(response.getQuery())
                    self.session.connection.connection.send(response)
                    self.__state_changed(State.TRANSFERRING)
                    raise nbxmpp.NodeProcessed
            else:
                args = {'candError': True}
                self.__state_changed(State.CAND_RECEIVED, args)
            return
        if cand_used:
            streamhost_cid = cand_used.getAttr('cid')
            streamhost_used = None
            for cand in self.transport.candidates:
                if cand['candidate_id'] == streamhost_cid:
                    streamhost_used = cand
                    break
            if streamhost_used is None or streamhost_used['type'] == 'proxy':
                if (app.socks5queue.listener and
                        not app.socks5queue.listener.connections):
                    app.socks5queue.listener.disconnect()
        if content.getTag('transport').getTag('activated'):
            self.state = State.TRANSFERRING
            app.socks5queue.send_file(self.file_props,
                                      self.session.connection.name, 'client')
            return
        args = {'content': content,
                'sendCand': False}
        if self.state == State.CAND_SENT:
            self.__state_changed(State.CAND_SENT_AND_RECEIVED, args)
            self.__state_changed(State.TRANSFERRING)
            raise nbxmpp.NodeProcessed
        self.__state_changed(State.CAND_RECEIVED, args)

    def __on_iq_result(self,
                       stanza: nbxmpp.Node,
                       content: nbxmpp.Node,
                       error: nbxmpp.Node | None,
                       action: str
                       ) -> None:
        log.info('__on_iq_result')

        if self.state in (State.NOT_STARTED, State.CAND_RECEIVED):
            self.__state_changed(State.INITIALIZED)
        elif self.state == State.CAND_SENT_AND_RECEIVED:
            if (not self.nominated_cand['our-cand'] and
                    not self.nominated_cand['peer-cand']):
                if not self.weinitiate:
                    return
                self.__state_changed(State.TRANSPORT_REPLACE)
                return
            # initiate transfer
            self.__state_changed(State.TRANSFERRING)

    def __transport_setup(self,
                          stanza: nbxmpp.Node,
                          content: nbxmpp.Node,
                          error: nbxmpp.Node | None,
                          action: str
                          ) -> None:
        # Sets up a few transport specific things for the file transfer
        if self.transport.type_ == TransportType.IBB:
            # No action required, just set the state to transferring
            self.state = State.TRANSFERRING
        else:
            self._listen_host()

    def on_connect(self, streamhost):
        '''
        send candidate-used stanza
        '''
        log.info('send_candidate_used')
        if streamhost is None:
            return
        args = {'streamhost': streamhost,
                'sendCand': True}
        self.nominated_cand['our-cand'] = streamhost
        self.__send_candidate(args)

    def _on_connect_error(self, sid: str) -> None:
        log.info('connect error, sid=%s', sid)
        args = {'candError': True,
                'sendCand': True}
        self.__send_candidate(args)

    def __send_candidate(self, args: dict[str, Any]) -> None:
        if self.state == State.CAND_RECEIVED:
            self.__state_changed(State.CAND_SENT_AND_RECEIVED, args)
        else:
            self.__state_changed(State.CAND_SENT, args)

    def _store_socks5_sid(self, sid: str, hash_id: str) -> None:
        # callback from socsk5queue.start_listener
        self.file_props.hash_ = hash_id

    def _listen_host(self) -> None:
        receiver = self.file_props.receiver
        sender = self.file_props.sender
        sha_str = helpers.get_auth_sha(self.file_props.sid, sender,
                                       receiver)
        self.file_props.sha_str = sha_str
        port = app.settings.get('file_transfers_port')
        listener = app.socks5queue.start_listener(
            port,
            sha_str,
            self._store_socks5_sid,
            self.file_props,
            typ='sender' if self.weinitiate else 'receiver')
        if not listener:
            # send error message, notify the user
            return

    def is_our_candidate_used(self) -> bool:
        '''
        If this method returns true then the candidate we nominated will be
        used, if false, the candidate nominated by peer will be used
        '''

        if not self.nominated_cand['peer-cand']:
            return True
        if not self.nominated_cand['our-cand']:
            return False
        peer_pr = int(self.nominated_cand['peer-cand']['priority'])
        our_pr = int(self.nominated_cand['our-cand']['priority'])
        if peer_pr != our_pr:
            return our_pr > peer_pr
        return self.weinitiate

    def start_ibb_transfer(self) -> None:
        if self.file_props.type_ == 's':
            self.__state_changed(State.TRANSFERRING)


def get_content(desc) -> JingleFileTransfer:
    return JingleFileTransfer


contents[Namespace.JINGLE_FILE_TRANSFER_5] = get_content



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import AvatarSize

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string


class SideBarListBox(Gtk.ListBox):
    __gtype_name__ = "SideBarListBox"

    def __init__(self) -> None:
        Gtk.ListBox.__init__(self)
        self.set_name("SideBarListBox")


@Gtk.Template(string=get_ui_string("side_bar_listbox_row.ui"))
class SideBarListBoxRow(Gtk.ListBoxRow, SignalManager):
    __gtype_name__ = "SideBarListBoxRow"

    _overlay: Gtk.Overlay = Gtk.Template.Child()
    _image: Gtk.Image = Gtk.Template.Child()
    _label: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        self._icon_name: str = ""
        self._item_id: str = ""
        self._index = 0
        self._drag_hotspot_x: float = 0
        self._drag_hotspot_y: float = 0
        Gtk.ListBoxRow.__init__(self)
        SignalManager.__init__(self)

    def do_unroot(self) -> None:
        Gtk.ListBoxRow.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def enable_as_drag_source(self) -> None:
        drag_source = Gtk.DragSource(actions=Gdk.DragAction.MOVE)
        self._connect(drag_source, "prepare", self._on_prepare)
        self._connect(drag_source, "drag-begin", self._on_drag_begin)
        self._connect(drag_source, "drag-end", self._on_drag_end)
        self.add_controller(drag_source)

    def _on_prepare(
        self, _drag_source: Gtk.DragSource, x: float, y: float
    ) -> Gdk.ContentProvider:
        self._drag_hotspot_x = x
        self._drag_hotspot_y = y

        value = GObject.Value()
        value.init(SideBarListBoxRow)
        value.set_object(self)

        return Gdk.ContentProvider.new_for_value(value)

    def _on_drag_begin(self, _drag_source: Gtk.DragSource, drag: Gdk.Drag) -> None:
        texture = app.app.avatar_storage.get_workspace_texture(
            self._item_id, AvatarSize.WORKSPACE, 1
        )
        if texture is not None:
            Gtk.DragIcon.set_from_paintable(
                drag, texture, int(self._drag_hotspot_x), int(self._drag_hotspot_y)
            )

        app.window.highlight_dnd_targets(self, True)

    def _on_drag_end(
        self, _drag_source: Gtk.DragSource, _drag: Gdk.Drag, _delete_data: bool
    ) -> None:
        app.window.highlight_dnd_targets(self, False)

    @GObject.Property(type=int)
    def index(self) -> int:  # pyright: ignore
        return self._index

    @index.setter
    def index(self, index: int) -> None:
        self._index = index

    @GObject.Property(type=str)
    def item_id(self) -> str:  # pyright: ignore
        return self._item_id

    @item_id.setter
    def item_id(self, item_id: str) -> None:
        self._item_id = item_id

    @GObject.Property(type=str)
    def icon_name(self) -> str:  # pyright: ignore
        return self._icon_name

    @icon_name.setter
    def icon_name(self, icon_name: str) -> None:
        self._icon_name = icon_name
        self._image.set_from_icon_name(icon_name)

    def set_workspace_id(self, workspace_id: str) -> None:
        self.item_id = workspace_id
        self.add_css_class("workspace-listbox-row")
        self.enable_as_drag_source()

    def set_from_paintable(self, paintable: Gdk.Paintable | None) -> None:
        self._image.set_from_paintable(paintable)

    def set_unread_notify(self, obj: GObject.Object) -> None:
        def _on_notify(obj_: GObject.Object, _param: GObject.ParamSpec) -> None:
            self.set_unread_count(obj_.unread_count)  # pyright: ignore

        self._connect(obj, "notify::unread-count", _on_notify)

    def set_unread_count(self, count: int) -> None:
        self._label.set_visible(count != 0)
        if count == 0:
            return

        if count < 1000:
            self._label.set_text(str(count))
        else:
            self._label.set_text("999+")

    def set_secondary_callback(self, callback: Any) -> None:
        controller = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(controller, "pressed", callback, self)
        self.add_controller(controller)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.modules.vcard4 import EmailProperty
from nbxmpp.modules.vcard4 import OrgProperty
from nbxmpp.modules.vcard4 import RoleProperty
from nbxmpp.modules.vcard4 import TelProperty
from nbxmpp.modules.vcard4 import TzProperty
from nbxmpp.modules.vcard4 import VCard

from gajim.common import app
from gajim.common import ged
from gajim.common.const import AvatarSize
from gajim.common.events import VCard4Received
from gajim.common.modules.contacts import BareContact
from gajim.common.util.decorators import event_filter
from gajim.common.util.user_strings import get_time_zone_string

from gajim.gtk.menus import get_account_menu
from gajim.gtk.preference.widgets import CopyButton
from gajim.gtk.status_message_row import StatusMessageSelectorRow
from gajim.gtk.status_selector import StatusSelector
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string


@Gtk.Template(string=get_ui_string("account_page.ui"))
class AccountPage(Gtk.Box, SignalManager):
    __gtype_name__ = "AccountPage"

    _avatar_image: Gtk.Image = Gtk.Template.Child()
    _name_label: Gtk.Label = Gtk.Template.Child()
    _account_label: Gtk.Label = Gtk.Template.Child()

    _settings_button: Gtk.Button = Gtk.Template.Child()
    _menu_button: Gtk.MenuButton = Gtk.Template.Child()

    _our_jid_row: Adw.ActionRow = Gtk.Template.Child()

    _status_selector: StatusSelector = Gtk.Template.Child()
    _status_message_row: StatusMessageSelectorRow = Gtk.Template.Child()

    _profile_button: Gtk.Button = Gtk.Template.Child()
    _profile_name_row: Adw.ActionRow = Gtk.Template.Child()
    _profile_org_row: Adw.ActionRow = Gtk.Template.Child()
    _profile_role_row: Adw.ActionRow = Gtk.Template.Child()
    _profile_email_row: Adw.ActionRow = Gtk.Template.Child()
    _profile_email_clipboard_button: CopyButton = Gtk.Template.Child()
    _profile_tel_row: Adw.ActionRow = Gtk.Template.Child()
    _profile_tel_clipboard_button: CopyButton = Gtk.Template.Child()
    _profile_timezone_row: Adw.ActionRow = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._account: str | None = None
        self._contact: BareContact | None = None

        self._menu_button.set_create_popup_func(self._on_menu_popup)

        self._profile_email = None
        self._profile_tel = None

        self._connect(
            self._profile_email_clipboard_button,
            "clicked",
            self._on_profile_email_copy_clicked,
        )
        self._connect(
            self._profile_tel_clipboard_button,
            "clicked",
            self._on_profile_tel_copy_clicked,
        )

        app.ged.register_event_handler(
            "vcard4-received", ged.GUI2, self._on_vcard_received
        )

    def set_account(self, account: str | None) -> None:
        if self._account == account:
            return

        self._disconnect_signals()

        self._account = account
        self._contact = None
        if account is not None:
            client = app.get_client(account)
            jid = client.get_own_jid().bare
            contact = client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            self._contact = contact

            self._settings_button.set_action_target_value(GLib.Variant("s", account))

            self._profile_button.set_action_target_value(GLib.Variant("s", account))
            self._profile_button.set_action_name(f"app.{account}-profile")

            vcard = client.get_module("VCard4").get_own_vcard()
            self._set_vcard_rows(vcard)

        self._connect_signals()

        self._update_page()

    def get_account(self) -> str | None:
        return self._account

    def _connect_signals(self) -> None:
        if self._account is None:
            return

        assert self._contact is not None

        self._contact.connect("avatar-update", self._on_avatar_update)
        app.settings.connect_signal(
            "account_label", self._on_account_label_changed, self._account
        )

    def _disconnect_signals(self) -> None:
        if self._account is None:
            return

        assert self._contact is not None

        self._contact.disconnect_all_from_obj(self)
        app.settings.disconnect_signals(self)

    def _on_profile_email_copy_clicked(self, _button: CopyButton) -> None:
        app.window.get_clipboard().set(self._profile_email)

    def _on_profile_tel_copy_clicked(self, _button: CopyButton) -> None:
        app.window.get_clipboard().set(self._profile_tel)

    def _on_menu_popup(self, menu_button: Gtk.MenuButton) -> None:
        assert self._account is not None
        menu_button.set_menu_model(get_account_menu(self._account))

    def _on_account_label_changed(self, value: str, *args: Any) -> None:
        self._update_account_label(value)

    def _on_avatar_update(self, *args: Any) -> None:
        self._update_avatar()

    def _update_account_label(self, label: str) -> None:
        self._account_label.set_text(label)

    def _update_avatar(self) -> None:
        assert self._contact is not None
        texture = self._contact.get_avatar(
            AvatarSize.ACCOUNT_PAGE, self.get_scale_factor(), add_show=False
        )
        self._avatar_image.set_from_paintable(texture)

    @event_filter(["account"])
    def _on_vcard_received(self, event: VCard4Received) -> None:
        self._set_vcard_rows(event.vcard)

    def _set_vcard_rows(self, vcard: VCard) -> None:
        self._profile_org_row.set_visible(False)
        self._profile_role_row.set_visible(False)
        self._profile_email_row.set_visible(False)
        self._profile_tel_row.set_visible(False)
        self._profile_timezone_row.set_visible(False)

        for prop in vcard.get_properties():
            match prop:
                case OrgProperty():
                    self._profile_org_row.set_subtitle(prop.values[0])
                    self._profile_org_row.set_visible(True)
                case RoleProperty():
                    self._profile_role_row.set_subtitle(prop.value)
                    self._profile_role_row.set_visible(True)
                case EmailProperty():
                    self._profile_email_row.set_subtitle(prop.value)
                    self._profile_email_row.set_action_target_value(
                        GLib.Variant("s", f"mailto:{prop.value}")
                    )
                    self._profile_email = prop.value
                    self._profile_email_row.set_visible(True)
                case TelProperty():
                    self._profile_tel_row.set_subtitle(prop.value)
                    self._profile_tel_row.set_action_target_value(
                        GLib.Variant("s", f"tel:{prop.value}")
                    )
                    self._profile_tel = prop.value
                    self._profile_tel_row.set_visible(True)
                case TzProperty():
                    self._profile_timezone_row.set_subtitle(get_time_zone_string(prop))
                    self._profile_timezone_row.set_visible(True)
                case _:
                    pass

    def _update_page(self) -> None:
        self._menu_button.set_sensitive(self._account is not None)

        if self._account is None:
            self._our_jid_row.set_subtitle("")
            self._account_label.set_text("")
            self._avatar_image.set_from_paintable(None)
            self._status_selector.set_account(None)
            self._status_message_row.set_account(None)
            return

        assert self._contact is not None

        account_label = app.settings.get_account_setting(self._account, "account_label")

        self._update_avatar()
        self._update_account_label(account_label)
        self._name_label.set_label(app.nicks[self._account])

        self._our_jid_row.set_subtitle(str(self._contact.jid))
        self._status_selector.set_account(self._account)
        self._status_message_row.set_account(self._account)

        self._profile_name_row.set_subtitle(app.nicks[self._account])



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


# Handles Jingle sessions (XEP 0166)

# TODO:
# * 'senders' attribute of 'content' element
# * security preconditions
# * actions:
#   - content-modify
#   - session-info
#   - security-info
#   - transport-accept, transport-reject
#   - Tie-breaking
# * timeout

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import logging
import uuid
from collections.abc import Callable
from enum import Enum
from enum import unique

import nbxmpp
from nbxmpp import JID
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.util import generate_id

from gajim.common import app
from gajim.common import events
from gajim.common.client import Client
from gajim.common.file_props import FilesProp
from gajim.common.jingle_content import get_jingle_content
from gajim.common.jingle_content import JingleContent
from gajim.common.jingle_content import JingleContentSetupException
from gajim.common.jingle_ft import State
from gajim.common.jingle_transport import get_jingle_transport
from gajim.common.jingle_transport import JingleTransportIBB
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.util.datetime import utc_now

if TYPE_CHECKING:
    from gajim.common.jingle_transport import JingleTransport


log = logging.getLogger('app.c.jingle_session')


JINGLE_EVENTS = {
    'jingle-connected-received': events.JingleConnectedReceived,
    'jingle-disconnected-received': events.JingleDisconnectedReceived,
    'jingle-request-received': events.JingleRequestReceived,
    'jingle-ft-cancelled-received': events.JingleFtCancelledReceived,
    'jingle-error-received': events.JingleErrorReceived
}


# FIXME: Move it to JingleSession.States?
@unique
class JingleStates(Enum):
    '''
    States in which jingle session may exist
    '''
    ENDED = 0
    PENDING = 1
    ACTIVE = 2


class OutOfOrder(Exception):
    '''
    Exception that should be raised when an action is received when in the wrong
    state
    '''


class TieBreak(Exception):
    '''
    Exception that should be raised in case of a tie, when we overrule the other
    action
    '''


class FailedApplication(Exception):
    '''
    Exception that should be raised in case responder supports none of the
    payload-types offered by the initiator
    '''


class JingleSession:
    '''
    This represents one jingle session, that is, one or more content types
    negotiated between an initiator and a responder.
    '''

    def __init__(self,
                 con: Client,
                 weinitiate: bool,
                 jid: str,
                 iq_id: str | None = None,
                 sid: str | None = None,
                 werequest: bool = False
                 ) -> None:
        '''
        con -- connection object,
        weinitiate -- boolean, are we the initiator?
        jid - jid of the other entity
        '''
        # negotiated contents
        self.contents: dict[tuple[str, str], JingleContent] = {}
        self.connection = con  # connection to use
        # our full jid
        self.ourjid = str(self.connection.get_own_jid())
        self.peerjid = jid  # jid we connect to
        # jid we use as the initiator
        self.initiator = self.ourjid if weinitiate else self.peerjid
        # jid we use as the responder
        self.responder = self.peerjid if weinitiate else self.ourjid
        # are we an initiator?
        self.weinitiate = weinitiate
        # Are we requesting or offering a file?
        self.werequest = werequest
        self.request = False
        # what state is session in? (one from JingleStates)
        self.state = JingleStates.ENDED
        if not sid:
            sid = generate_id()
        self.sid = sid  # sessionid
        # iq stanza id, used to determine which sessions to summon callback
        # later on when iq-result stanza arrives
        if iq_id is not None:
            self.iq_ids = [iq_id]
        else:
            self.iq_ids = []
        self.accepted = True  # is this session accepted by user
        # Tells whether this session is a file transfer or not
        self.session_type_ft = False
        # callbacks to call on proper contents
        # use .prepend() to add new callbacks, especially when you're going
        # to send error instead of ack
        self.callbacks: dict[str, list[Callable[
            [nbxmpp.Node, nbxmpp.Node, nbxmpp.Node | None, str],
            None]]] = {
            'content-accept': [self.__ack,
                               self.__on_content_accept,
                               self.__broadcast],
            'content-add': [self.__ack,
                            self.__on_content_add,
                            self.__broadcast],  # TODO
            'content-modify': [self.__ack],  # TODO
            'content-reject': [self.__ack, self.__on_content_remove],
            'content-remove': [self.__ack, self.__on_content_remove],
            'description-info': [self.__ack, self.__broadcast],  # TODO
            'security-info': [self.__ack],  # TODO
            'session-accept': [self.__ack, self.__on_session_accept,
                               self.__on_content_accept,
                               self.__broadcast],
            'session-info': [self.__ack, self.__broadcast,
                             self.__on_session_info],
            'session-initiate': [self.__ack, self.__on_session_initiate,
                                 self.__broadcast],
            'session-terminate': [self.__ack, self.__on_session_terminate,
                                  self.__broadcast_all],
            'transport-info': [self.__ack, self.__broadcast],
            'transport-replace': [self.__ack, self.__broadcast,
                                  self.__on_transport_replace],  # TODO
            'transport-accept': [self.__ack, self.__on_session_accept,
                                 self.__on_content_accept,
                                 self.__broadcast],
            'transport-reject': [self.__ack],  # TODO
            'iq-result': [self.__broadcast],
            'iq-error': [self.__on_error],
        }

    def collect_iq_id(self, iq_id: str) -> None:
        if iq_id is not None:
            self.iq_ids.append(iq_id)

    def approve_session(self) -> None:
        '''
        Called when user accepts session in UI (when we aren't the initiator)
        '''
        self.accept_session()

    def decline_session(self) -> None:
        '''
        Called when user declines session in UI (when we aren't the initiator)
        '''
        reason = nbxmpp.Node('reason')
        reason.addChild('decline')
        self._session_terminate(reason)

    def cancel_session(self) -> None:
        '''
        Called when user declines session in UI (when we aren't the initiator)
        '''
        reason = nbxmpp.Node('reason')
        reason.addChild('cancel')
        self._session_terminate(reason)

    def approve_content(self, media: str, name: str | None = None) -> None:
        content = self.get_content(media, name)
        if content:
            content.accepted = True
            self.on_session_state_changed(content)

    def reject_content(self, media: str, name: str | None = None) -> None:
        content = self.get_content(media, name)
        if content:
            if self.state == JingleStates.ACTIVE:
                self.__content_reject(content)
            content.destroy()
            self.on_session_state_changed()

    def end_session(self) -> None:
        '''
        Called when user stops or cancel session in UI
        '''
        reason = nbxmpp.Node('reason')
        if self.state == JingleStates.ACTIVE:
            reason.addChild('success')
        else:
            reason.addChild('cancel')
        self._session_terminate(reason)

    def get_content(self,
                    media: str | None = None,
                    name: str | None = None
                    ) -> JingleContent | None:
        if media is None:
            return None
        for content in self.contents.values():
            if content.media == media:
                if name is None or content.name == name:
                    return content
        return None

    def add_content(self,
                    name: str,
                    content: JingleContent,
                    creator: str = 'we'
                    ) -> None:
        '''
        Add new content to session. If the session is active, this will send
        proper stanza to update session

        Creator must be one of ('we', 'peer', 'initiator', 'responder')
        '''
        assert creator in ('we', 'peer', 'initiator', 'responder')
        if ((creator == 'we' and self.weinitiate) or
                (creator == 'peer' and not self.weinitiate)):
            creator = 'initiator'
        elif ((creator == 'peer' and self.weinitiate) or
                (creator == 'we' and not self.weinitiate)):
            creator = 'responder'
        content.creator = creator
        content.name = name
        self.contents[(creator, name)] = content
        if (creator == 'initiator') == self.weinitiate:
            # The content is from us, accept it
            content.accepted = True

    def remove_content(self,
                       creator: str,
                       name: str,
                       reason: nbxmpp.Node | None = None
                       ) -> None:
        '''
        Remove the content `name` created by `creator`
        by sending content-remove, or by sending session-terminate if
        there is no content left.
        '''
        if (creator, name) in self.contents:
            content = self.contents[(creator, name)]
            self.__content_remove(content, reason)
            self.contents[(creator, name)].destroy()
        if not self.contents:
            self.end_session()

    def modify_content(self,
                       creator: str,
                       name: str,
                       transport: JingleTransport | None = None
                       ) -> None:
        '''
        Currently used for transport replacement
        '''
        content = self.contents[(creator, name)]
        file_props = content.transport.file_props
        file_props.transport_sid = transport.sid
        transport.set_file_props(file_props)
        content.transport = transport
        # The content will have to be resend now that it is modified
        content.sent = False
        content.accepted = True

    def on_session_state_changed(self,
                                 content: JingleContent | None = None
                                 ) -> None:
        if self.state == JingleStates.ENDED:
            # Session not yet started, only one action possible:
            # session-initiate
            if self.is_ready() and self.weinitiate:
                self.__session_initiate()
        elif self.state == JingleStates.PENDING:
            # We can either send a session-accept or a content-add
            if self.is_ready() and not self.weinitiate:
                self.__session_accept()
            elif (content and
                  (content.creator == 'initiator') == self.weinitiate):
                self.__content_add(content)
            elif content and self.weinitiate:
                self.__content_accept(content)
        elif self.state == JingleStates.ACTIVE:
            # We can either send a content-add or a content-accept. However, if
            # we are sending a file we can only use session_initiate.
            if not content:
                return
            is_initiator = content.creator == 'initiator'
            we_created_content = is_initiator == self.weinitiate
            if we_created_content and content.media == 'file':
                self.__session_initiate()
            if we_created_content:
                # We initiated this content. It's a pending content-add.
                self.__content_add(content)
            else:
                # The other side created this content, we accept it.
                self.__content_accept(content)

    def is_ready(self) -> bool:
        '''
        Return True when all codecs and candidates are ready (for all contents)
        '''
        ready = [content.is_ready() for content in self.contents.values()]
        return all(ready) and self.accepted

    def accept_session(self) -> None:
        '''
        Mark the session as accepted
        '''
        self.accepted = True
        self.on_session_state_changed()

    def start_session(self) -> None:
        '''
        Mark the session as ready to be started
        '''
        self.accepted = True
        self.on_session_state_changed()

    def send_session_info(self) -> Any:
        pass

    def send_content_accept(self, content: nbxmpp.Node) -> None:
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('content-accept')
        jingle.addChild(node=content)
        self.connection.connection.send(stanza)

    def send_transport_info(self, content: nbxmpp.Node) -> None:
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('transport-info')
        jingle.addChild(node=content)
        self.connection.connection.send(stanza)
        self.collect_iq_id(stanza.getID())

    def send_description_info(self, content: nbxmpp.Node) -> None:
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('description-info')
        jingle.addChild(node=content)
        self.connection.connection.send(stanza)

    def on_stanza(self, stanza: nbxmpp.Node) -> None:
        '''
        A callback for ConnectionJingle. It gets stanza, then tries to send
        it to all internally registered callbacks. First one to raise
        nbxmpp.NodeProcessed breaks function
        '''
        jingle = stanza.getTag('jingle')
        error = stanza.getTag('error')
        if error:
            # it's an iq-error stanza
            action = 'iq-error'
        elif jingle:
            # it's a jingle action
            action = jingle.getAttr('action')
            if action not in self.callbacks:
                self.__send_error(stanza, 'bad-request')
                return
            # FIXME: If we aren't initiated and it's not a session-initiate...
            if action not in ['session-initiate', 'session-terminate'] \
                    and self.state == JingleStates.ENDED:
                self.__send_error(stanza, 'item-not-found', 'unknown-session')
                return
        else:
            # it's an iq-result (ack) stanza
            action = 'iq-result'
        callables = self.callbacks[action]
        try:
            for call in callables:
                call(stanza=stanza, jingle=jingle, error=error, action=action)
        except nbxmpp.NodeProcessed:
            pass
        except TieBreak:
            self.__send_error(stanza, 'conflict', 'tiebreak')
        except OutOfOrder:
            # FIXME
            self.__send_error(stanza, 'unexpected-request', 'out-of-order')
        except FailedApplication:
            reason = nbxmpp.Node('reason')
            reason.addChild('failed-application')
            self._session_terminate(reason)

    def __ack(self,
              stanza: nbxmpp.Node,
              jingle: nbxmpp.Node,
              error: nbxmpp.Node | None,
              action: str
              ) -> None:
        '''
        Default callback for action stanzas -- simple ack and stop processing
        '''
        response = stanza.buildReply('result')
        response.delChild(response.getQuery())
        self.connection.connection.send(response)

    def __on_error(self,
                   stanza: nbxmpp.Node,
                   jingle: nbxmpp.Node,
                   error: nbxmpp.Node,
                   action: str
                   ) -> None:
        # FIXME
        text = error.getTagData('text')
        error_name = None
        for child in error.getChildren():
            if child.getNamespace() == Namespace.JINGLE_ERRORS:
                error_name = child.getName()
                break
            if child.getNamespace() == Namespace.STANZAS:
                error_name = child.getName()
        self.__dispatch_error(error_name, text, error.getAttr('type'))

    def transport_replace(self) -> None:
        transport = JingleTransportIBB()
        # For debug only, delete this and replace for a function
        # that will identify contents by its sid
        for creator, name in self.contents:
            self.modify_content(creator, name, transport)
            cont = self.contents[(creator, name)]
            cont.transport = transport
        stanza, jingle = self.__make_jingle('transport-replace')
        self.__append_contents(jingle)
        self.__broadcast(stanza, jingle, None, 'transport-replace')
        self.connection.connection.send(stanza)
        self.state = JingleStates.PENDING

    def __on_transport_replace(self,
                               stanza: nbxmpp.Node,
                               jingle: nbxmpp.Node,
                               error: nbxmpp.Node | None,
                               action: str
                               ) -> None:
        for content in jingle.iterTags('content'):
            creator = content['creator']
            name = content['name']
            transport_ns = content.getTag('transport').getNamespace()
            if (creator, name) in self.contents:
                if transport_ns == Namespace.JINGLE_ICE_UDP:
                    # FIXME: We don't manage anything else than ICE-UDP now...
                    # What was the previous transport?!?
                    # Anyway, content's transport is not modifiable yet
                    pass
                elif transport_ns == Namespace.JINGLE_IBB:
                    transport = JingleTransportIBB(node=content.getTag(
                        'transport'))
                    self.modify_content(creator, name, transport)
                    self.state = JingleStates.PENDING
                    self.contents[(creator, name)].state = \
                        State.TRANSPORT_REPLACE
                    self.__ack(stanza, jingle, error, action)
                    self.__transport_accept(transport)
                else:
                    stanza, jingle = self.__make_jingle('transport-reject')
                    content = jingle.setTag('content',
                                            attrs={'creator': creator,
                                                   'name': name})
                    content.setTag('transport', namespace=transport_ns)
                    self.connection.connection.send(stanza)
                    raise nbxmpp.NodeProcessed
            else:
                # FIXME: This resource is unknown to us, what should we do?
                # For now, reject the transport
                stanza, jingle = self.__make_jingle('transport-reject')
                content = jingle.setTag('content', attrs={'creator': creator,
                                                          'name': name})
                content.setTag('transport', namespace=transport_ns)
                self.connection.connection.send(stanza)
                raise nbxmpp.NodeProcessed

    def __on_session_info(self,
                          stanza: nbxmpp.Node,
                          jingle: nbxmpp.Node,
                          error: nbxmpp.Node | None,
                          action: str
                          ) -> None:
        # TODO: active, (un)hold, (un)mute
        ringing = jingle.getTag('ringing')
        if ringing is not None:
            # ignore ringing
            raise nbxmpp.NodeProcessed
        if self.state != JingleStates.ACTIVE:
            raise OutOfOrder
        for child in jingle.getChildren():
            if child.getName() == 'checksum':
                hash_ = child.getTag('file').getTag(
                    name='hash',
                    namespace=Namespace.HASHES_2)
                if hash_ is None:
                    continue
                algo = hash_.getAttr('algo')
                if algo in nbxmpp.Hashes2.supported:
                    file_props = FilesProp.getFileProp(self.connection.name,
                                                       self.sid)
                    file_props.algo = algo
                    file_props.hash_ = hash_.getData()
                    raise nbxmpp.NodeProcessed
        self.__send_error(stanza, 'feature-not-implemented', 'unsupported-info',
                          type_='modify')
        raise nbxmpp.NodeProcessed

    def __on_content_remove(self,
                            stanza: nbxmpp.Node,
                            jingle: nbxmpp.Node,
                            error: nbxmpp.Node | None,
                            action: str
                            ) -> None:
        for content in jingle.iterTags('content'):
            creator = content['creator']
            name = content['name']
            if (creator, name) in self.contents:
                content = self.contents[(creator, name)]
                # TODO: this will fail if content is not an RTP content
                self._raise_event('jingle-disconnected-received',
                                  media=content.media,
                                  reason='removed')
                content.destroy()
        if not self.contents:
            reason = nbxmpp.Node('reason')
            reason.setTag('success')
            self._session_terminate(reason)

    def __on_session_accept(self,
                            stanza: nbxmpp.Node,
                            jingle: nbxmpp.Node,
                            error: nbxmpp.Node | None,
                            action: str
                            ) -> None:
        # FIXME
        if self.state != JingleStates.PENDING:
            raise OutOfOrder
        self.state = JingleStates.ACTIVE

    @staticmethod
    def __on_content_accept(stanza: nbxmpp.Node,
                            jingle: nbxmpp.Node,
                            error: nbxmpp.Node | None,
                            action: str
                            ) -> None:
        '''
        Called when we get content-accept stanza or equivalent one (like
        session-accept)
        '''
        # check which contents are accepted
        # for content in jingle.iterTags('content'):
        #     creator = content['creator']
        #     name = content['name']
        return

    def __on_content_add(self,
                         stanza: nbxmpp.Node,
                         jingle: nbxmpp.Node,
                         error: nbxmpp.Node | None,
                         action: str
                         ) -> None:
        if self.state == JingleStates.ENDED:
            raise OutOfOrder
        parse_result = self.__parse_contents(jingle)
        contents = parse_result[0]
        # rejected_contents = parse_result[1]
        # for name, creator in rejected_contents:
        #     content = JingleContent()
        #     self.add_content(name, content, creator)
        #     self.__content_reject(content)
        #     self.contents[(content.creator, content.name)].destroy()
        self._raise_event('jingle-request-received', contents=contents)

    def __on_session_initiate(self,
                              stanza: nbxmpp.Node,
                              jingle: nbxmpp.Node,
                              error: nbxmpp.Node | None,
                              action: str
                              ) -> None:
        '''
        We got a jingle session request from other entity, therefore we are the
        receiver... Unpack the data, inform the user
        '''
        if self.state != JingleStates.ENDED:
            raise OutOfOrder

        # In "session-initiate", the <jingle/> element SHOULD possess an
        # 'initiator' attribute. If it differs from the JID 'from', jingle
        # 'initiator' SHOULD be ignored.
        if jingle['initiator'] is None:
            self.initiator = stanza['from']
        elif jingle['initiator'] != stanza['from']:
            self.initiator = stanza['from']
        else:
            self.initiator = jingle['initiator']

        self.responder = self.ourjid
        self.peerjid = self.initiator
        self.accepted = False   # user did not accept this session yet
        # TODO: If the initiator is unknown to the receiver (e.g., via presence
        # subscription) and the receiver has a policy of not communicating via
        # Jingle with unknown entities, it SHOULD return a
        # <service-unavailable/> error.
        # Lets check what kind of jingle session does the peer want
        contents, _contents_rejected, reason_txt = self.__parse_contents(jingle)

        # If there's no content we understand...
        if not contents:
            # TODO: http://xmpp.org/extensions/xep-0166.html#session-terminate
            reason = nbxmpp.Node('reason')
            reason.setTag(reason_txt)
            self.__ack(stanza, jingle, error, action)
            self._session_terminate(reason)
            raise nbxmpp.NodeProcessed

        # If we are not receiving a file
        # Check if there's already a session with this user:
        if contents[0].media != 'file':
            jingle_module = self.connection.get_module('Jingle')
            for session in jingle_module.get_jingle_sessions(self.peerjid):
                if session is not self:
                    reason = nbxmpp.Node('reason')
                    alternative_session = reason.setTag('alternative-session')
                    alternative_session.setTagData('sid', session.sid)
                    self.__ack(stanza, jingle, error, action)
                    self._session_terminate(reason)
                    raise nbxmpp.NodeProcessed
        else:
            # Stop if we don't have the requested file or the peer is not
            # allowed to request the file
            request = contents[0].senders == 'responder'
            if request:
                self.request = True
                hash_tag = request.getTag('file').getTag('hash')
                hash_data = hash_tag.getData() if hash_tag else None
                n = request.getTag('file').getTag('name')
                n = n.getData() if n else None
                pjid = app.get_jid_without_resource(self.peerjid)
                file_info = self.connection.get_module('Jingle').get_file_info(
                    pjid, hash_data, n, self.connection.name)
                if not file_info:
                    log.warning('The peer %s is requesting a '
                                'file that we don’t have or '
                                'it is not allowed to request', pjid)
                    self.decline_session()
                    raise nbxmpp.NodeProcessed

        self.state = JingleStates.PENDING
        # Send event about starting a session
        self._raise_event('jingle-request-received', contents=contents)

        # Check if it's an A/V call and add it to the archive
        content_types: list[str] = []
        for item in contents:
            content_types.append(item.media)
        if not any(item in ('audio', 'video') for item in content_types):
            return

        account = self.connection.name
        jid = JID.from_string(self.peerjid)

        call_data = mod.Call(
            sid=self.sid,
            state=0,  # TODO
        )

        message = mod.Message(
            account_=account,
            remote_jid_=jid.new_as_bare(),
            resource=None,
            type=MessageType.CHAT,
            direction=ChatDirection.INCOMING,
            timestamp=utc_now(),
            state=MessageState.ACKNOWLEDGED,
            id=str(uuid.uuid4()),
            call=call_data,
        )

        app.storage.archive.insert_object(message)

    def __broadcast(self,
                    stanza: nbxmpp.Node,
                    jingle: nbxmpp.Node,
                    error: nbxmpp.Node | None,
                    action: str
                    ) -> None:
        '''
        Broadcast the stanza contents to proper content handlers
        '''
        # if jingle is None: # it is a iq-result stanza
        #    for cn in self.contents.values():
        #        cn.on_stanza(stanza, None, error, action)
        #    return
        # special case: iq-result stanza does not come with a jingle element
        if action == 'iq-result':
            for cn in self.contents.values():
                cn.on_stanza(stanza, None, error, action)
            return
        for content in jingle.iterTags('content'):
            name = content['name']
            creator = content['creator']
            if (creator, name) not in self.contents:
                text = f'Content {name} (created by {creator}) does not exist'
                self.__send_error(stanza,
                                  'bad-request',
                                  text=text,
                                  type_='modify')
                raise nbxmpp.NodeProcessed

            cn = self.contents[(creator, name)]
            cn.on_stanza(stanza, content, error, action)

    def __on_session_terminate(self,
                               stanza: nbxmpp.Node,
                               jingle: nbxmpp.Node,
                               error: nbxmpp.Node | None,
                               action: str
                               ) -> None:
        self.connection.get_module('Jingle').delete_jingle_session(self.sid)
        reason, text = self.__reason_from_stanza(jingle)
        if reason not in ('success', 'cancel', 'decline'):
            self.__dispatch_error(reason, text)
        if text:
            text = f'{reason} ({text})'
        else:
            # TODO
            text = reason
        if reason == 'decline':
            self._raise_event('jingle-disconnected-received',
                              media=None,
                              reason=text)
        if reason == 'success':
            self._raise_event('jingle-disconnected-received',
                              media=None,
                              reason=text)
        if reason == 'cancel' and self.session_type_ft:
            pass
            # TODO: Jingle FT
            # self._raise_event('jingle-ft-cancelled-received',
            #                   media=None,
            #                   reason=text)

    def __broadcast_all(self,
                        stanza: nbxmpp.Node,
                        jingle: nbxmpp.Node,
                        error: nbxmpp.Node | None,
                        action: str
                        ) -> None:
        '''
        Broadcast the stanza to all content handlers
        '''
        for content in self.contents.values():
            content.on_stanza(stanza, None, error, action)

    def __parse_contents(self,
                         jingle: nbxmpp.Node
                         ) -> tuple[
                             list[JingleContent],
                             list[tuple[str, str]],
                             str | None]:
        # TODO: Needs some reworking
        contents: list[JingleContent] = []
        contents_rejected: list[tuple[str, str]] = []
        reasons: set[str] = set()
        for element in jingle.iterTags('content'):
            transport = get_jingle_transport(element.getTag('transport'))
            if transport:
                transport.ourjid = self.ourjid
            content_type = get_jingle_content(element.getTag('description'))
            if content_type:
                try:
                    if transport:
                        content = content_type(self, transport=transport)
                        self.add_content(element['name'],
                                         content,
                                         'peer')
                        contents.append(content)
                    else:
                        reasons.add('unsupported-transports')
                        contents_rejected.append((element['name'], 'peer'))
                except JingleContentSetupException:
                    reasons.add('failed-application')
            else:
                contents_rejected.append((element['name'], 'peer'))
                reasons.add('unsupported-applications')
        failure_reason = None
        # Store the first reason of failure
        for reason in ('failed-application', 'unsupported-transports',
                       'unsupported-applications'):
            if reason in reasons:
                failure_reason = reason
                break
        return (contents, contents_rejected, failure_reason)

    def __dispatch_error(self,
                         error: str | None = None,
                         text: str | None = None,
                         type_: str | None = None
                         ) -> None:
        if text:
            text = '{error} ({text})'
        if type_ != 'modify':
            self._raise_event('jingle-error-received', reason=text or error)

    @staticmethod
    def __reason_from_stanza(stanza: nbxmpp.Node) -> tuple[str, str]:
        # TODO: Move to GUI?
        reason = 'success'
        reasons = [
            'success', 'busy', 'cancel', 'connectivity-error', 'decline',
            'expired', 'failed-application', 'failed-transport',
            'general-error', 'gone', 'incompatible-parameters', 'media-error',
            'security-error', 'timeout', 'unsupported-applications',
            'unsupported-transports'
        ]
        tag = stanza.getTag('reason')
        text = ''
        if tag:
            text = tag.getTagData('text')
            for r in reasons:
                if tag.getTag(r):
                    reason = r
                    break
        return (reason, text)

    def __make_jingle(self,
                      action: str,
                      reason: nbxmpp.Node | None = None
                      ) -> tuple[Iq, nbxmpp.Node]:
        stanza = nbxmpp.Iq(typ='set', to=nbxmpp.JID.from_string(self.peerjid),
                           frm=self.ourjid)
        attrs = {
            'action': action,
            'sid': self.sid,
            'initiator': self.initiator
        }
        jingle = stanza.addChild('jingle', attrs=attrs,
                                 namespace=Namespace.JINGLE)
        if reason is not None:
            jingle.addChild(node=reason)
        return stanza, jingle

    def __send_error(self,
                     stanza: nbxmpp.Node,
                     error: str | None,
                     jingle_error: str | None = None,
                     text: str | None = None,
                     type_: str | None = None
                     ) -> None:
        err_stanza = nbxmpp.Error(stanza, f'{Namespace.STANZAS} {error}')
        err = err_stanza.getTag('error')
        if type_:
            err.setAttr('type', type_)
        if jingle_error:
            err.setTag(jingle_error, namespace=Namespace.JINGLE_ERRORS)
        if text:
            err.setTagData('text', text)
        self.connection.connection.send(err_stanza)
        self.__dispatch_error(jingle_error or error, text, type_)

    @staticmethod
    def __append_content(jingle: nbxmpp.Node, content: JingleContent) -> None:
        '''
        Append <content/> element to <jingle/> element
        '''
        jingle.addChild('content',
                        attrs={'name': content.name,
                               'creator': content.creator,
                               'senders': content.senders})

    def __append_contents(self, jingle: nbxmpp.Node) -> None:
        '''
        Append all <content/> elements to <jingle/>
        '''
        # TODO: integrate with __appendContent?
        # TODO: parameters 'name', 'content'?
        for content in self.contents.values():
            if content.is_ready():
                self.__append_content(jingle, content)

    def __session_initiate(self) -> None:
        assert self.state == JingleStates.ENDED
        stanza, jingle = self.__make_jingle('session-initiate')
        self.__append_contents(jingle)
        self.__broadcast(stanza, jingle, None, 'session-initiate-sent')
        self.connection.connection.send(stanza)
        self.collect_iq_id(stanza.getID())
        self.state = JingleStates.PENDING

    def __session_accept(self) -> None:
        assert self.state == JingleStates.PENDING
        stanza, jingle = self.__make_jingle('session-accept')
        self.__append_contents(jingle)
        self.__broadcast(stanza, jingle, None, 'session-accept-sent')
        self.connection.connection.send(stanza)
        self.collect_iq_id(stanza.getID())
        self.state = JingleStates.ACTIVE

    def __session_info(self, payload: nbxmpp.Node | None = None) -> None:
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('session-info')
        if payload:
            jingle.addChild(node=payload)
        self.connection.connection.send(stanza)

    def _JingleFileTransfer__session_info(self, payload: nbxmpp.Node) -> None:
        # For some strange reason when I call
        # self.session.__session_info(payload) from the jingleFileTransfer
        # object within a thread, this method gets called instead. Even though,
        # it isn't being called explicitly.
        self.__session_info(payload)

    def _session_terminate(self, reason: nbxmpp.Node | None = None) -> None:
        stanza, jingle = self.__make_jingle('session-terminate', reason=reason)
        self.__broadcast_all(stanza, jingle, None, 'session-terminate-sent')
        if self.connection.connection and self.connection.state.is_available:
            self.connection.connection.send(stanza)
        # TODO: Move to GUI?
        reason_text, text = self.__reason_from_stanza(jingle)
        if reason not in ('success', 'cancel', 'decline'):
            self.__dispatch_error(reason_text, text)
        if text:
            text = f'{reason_text} ({text})'
        else:
            text = reason_text
        self.connection.get_module('Jingle').delete_jingle_session(self.sid)
        self._raise_event('jingle-disconnected-received',
                          media=None,
                          reason=text)

    def __content_add(self, content: JingleContent) -> None:
        # TODO: test
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('content-add')
        self.__append_content(jingle, content)
        self.__broadcast(stanza, jingle, None, 'content-add-sent')
        id_ = self.connection.connection.send(stanza)
        self.collect_iq_id(id_)

    def __content_accept(self, content: JingleContent) -> None:
        # TODO: test
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('content-accept')
        self.__append_content(jingle, content)
        self.__broadcast(stanza, jingle, None, 'content-accept-sent')
        id_ = self.connection.connection.send(stanza)
        self.collect_iq_id(id_)

    def __content_reject(self, content: JingleContent) -> None:
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('content-reject')
        self.__append_content(jingle, content)
        self.connection.connection.send(stanza)
        # TODO: this will fail if content is not an RTP content
        self._raise_event('jingle-disconnected-received',
                          media=content.media,
                          reason='rejected')

    def __content_remove(self,
                         content: JingleContent,
                         reason: nbxmpp.Node | None = None
                         ) -> None:
        assert self.state != JingleStates.ENDED
        if self.connection.connection and self.connection.state.is_available:
            stanza, jingle = self.__make_jingle('content-remove', reason=reason)
            self.__append_content(jingle, content)
            self.connection.connection.send(stanza)
        # TODO: this will fail if content is not an RTP content
        self._raise_event('jingle-disconnected-received',
                          media=content.media,
                          reason='removed')

    def content_negotiated(self, media: str) -> None:
        self._raise_event('jingle-connected-received', media=media)

    def __transport_accept(self, transport: JingleTransport) -> None:
        assert self.state != JingleStates.ENDED
        stanza, jingle = self.__make_jingle('transport-accept')
        self.__append_contents(jingle)
        self.__broadcast(stanza, jingle, None, 'transport-accept')
        self.connection.connection.send(stanza)
        self.collect_iq_id(stanza.getID())
        self.state = JingleStates.ACTIVE

    def _raise_event(self, name: str, **kwargs: Any) -> None:
        jid, resource = app.get_room_and_nick_from_fjid(
            str(self.peerjid))
        jid = JID.from_string(jid)
        event_class = JINGLE_EVENTS[name]

        app.ged.raise_event(
            event_class(conn=self.connection,
                        account=self.connection.name,
                        fjid=self.peerjid,
                        jid=jid,
                        sid=self.sid,
                        resource=resource,
                        jingle_session=self,
                        **kwargs))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

import time

import nbxmpp
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import GtkSource
from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common import ged
from gajim.common.const import Direction
from gajim.common.events import AccountDisabled
from gajim.common.events import AccountEnabled
from gajim.common.events import StanzaReceived
from gajim.common.events import StanzaSent
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.logging_helpers import get_log_console_handler

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.const import Setting
from gajim.gtk.const import SettingKind
from gajim.gtk.const import SettingType
from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.settings import SettingsDialog
from gajim.gtk.util.misc import at_the_end
from gajim.gtk.util.misc import scroll_to
from gajim.gtk.util.styling import get_source_view_style_scheme
from gajim.gtk.window import GajimAppWindow

STANZA_PRESETS = {
    "Presence": (
        '<presence xmlns="jabber:client">\n'
        "<show></show>\n"
        "<status></status>\n"
        "<priority></priority>\n"
        "</presence>"
    ),
    "Message": (
        '<message to="" type="" xmlns="jabber:client">\n<body></body>\n</message>'
    ),
    "Iq": ('<iq to="" type="" xmlns="jabber:client">\n<query xmlns=""></query>\n</iq>'),
    "XEP-0030: Disco Info Query": (
        '<iq to="" type="get" xmlns="jabber:client">\n'
        f'<query xmlns="{Namespace.DISCO_INFO}">'
        "</query>\n</iq>"
    ),
    "XEP-0092: Software Version Query": (
        '<iq to="" type="get" xmlns="jabber:client">\n'
        f'<query xmlns="{Namespace.VERSION}">'
        "</query>\n</iq>"
    ),
}


class DebugConsoleWindow(GajimAppWindow, EventHelper):
    def __init__(self) -> None:
        GajimAppWindow.__init__(
            self,
            name="DebugConsoleWindow",
            default_width=800,
            default_height=600,
            header_bar=True,
        )

        EventHelper.__init__(self)

        self._selected_send_account: str | None = None
        self._filter_dialog: SettingsDialog | None = None
        self._sent_stanzas = SentSzanzas()
        self._last_selected_ts = 0

        self._presence = True
        self._message = True
        self._iq = True
        self._stream = False
        self._incoming = True
        self._outgoing = True

        self._ui = get_builder("debug_console.ui")
        header_bar = self.get_header_bar()
        assert header_bar is not None
        header_bar.set_title_widget(self._ui.header_box)
        self.set_child(self._ui.stack)

        self._ui.paned.set_position(self._ui.paned.get_property("max-position"))

        action = Gio.SimpleAction.new("clear-window")
        self._connect(action, "activate", self._on_clear_window)
        self.add_action(action)

        shortcut_controller = Gtk.ShortcutController()
        callback_action = Gtk.CallbackAction.new(self._on_clear_window)  # type: ignore
        shortcut_trigger = Gtk.ShortcutTrigger.parse_string("<Control>l")
        self._shortcut = Gtk.Shortcut.new(shortcut_trigger, callback_action)
        shortcut_controller.add_shortcut(self._shortcut)
        self.add_controller(shortcut_controller)

        accounts = self._get_accounts()
        self._account_dropdown: GajimDropDown[str] = GajimDropDown(
            fixed_width=15, data=accounts
        )
        if accounts:
            self._selected_send_account = next(iter(accounts))

        self._connect(
            self._account_dropdown, "notify::selected", self._on_account_change
        )
        self._account_dropdown.set_visible(False)
        self._ui.actionbox.append(self._account_dropdown)
        self._ui.actionbox.reorder_child_after(
            self._account_dropdown, self._ui.account_label
        )

        self._create_tags()
        self._add_stanza_presets()

        self._connect(
            self._ui.filter_options_button, "clicked", self._on_filter_options
        )
        self._connect(self._ui.paste, "clicked", self._on_paste_previous)
        self._connect(
            self._ui.stanza_presets_listbox, "row-activated", self._on_row_activated
        )
        self._connect(self._ui.search_entry, "activate", self._on_search_activate)
        self._connect(self._ui.search_forward, "clicked", self._on_search_clicked)
        self._connect(self._ui.search_backward, "clicked", self._on_search_clicked)
        self._connect(
            self._ui.jump_to_end_button, "clicked", self._on_jump_to_end_clicked
        )
        self._connect(self._ui.send, "clicked", self._on_send)
        self._connect(self._ui.edit_toggle, "toggled", self._on_input)
        self._connect(self._ui.search_toggle, "toggled", self._on_search_toggled)

        source_manager = GtkSource.LanguageManager.get_default()
        lang = source_manager.get_language("xml")
        self._ui.protocol_view.get_buffer().set_language(lang)
        self._ui.input_entry.get_buffer().set_language(lang)

        style_scheme = get_source_view_style_scheme()
        if style_scheme is not None:
            self._ui.protocol_view.get_buffer().set_style_scheme(style_scheme)
            self._ui.input_entry.get_buffer().set_style_scheme(style_scheme)
            self._ui.log_view.get_buffer().set_style_scheme(style_scheme)

        self._search_settings = GtkSource.SearchSettings(wrap_around=True)
        self._search_context = GtkSource.SearchContext.new(
            self._ui.protocol_view.get_buffer(), self._search_settings
        )

        for record in app.logging_records:
            self._add_log_record(record)

        log_handler = get_log_console_handler()
        log_handler.set_callback(self._add_log_record)

        self._connect(
            self._ui.stack, "notify::visible-child-name", self._on_stack_child_changed
        )

        vadjustment = self._ui.scrolled.get_vadjustment()
        self._connect(vadjustment, "notify::upper", self._on_adj_upper_changed)
        self._connect(vadjustment, "notify::value", self._on_adj_value_changed)

        controller = self.get_default_controller()
        self._connect(controller, "key-pressed", self._on_key_pressed)

        self.register_events(
            [
                ("stanza-received", ged.GUI1, self._on_stanza_received),
                ("stanza-sent", ged.GUI1, self._on_stanza_sent),
                ("account-enabled", ged.GUI1, self._on_account_changed),
                ("account-disabled", ged.GUI1, self._on_account_changed),
            ]
        )
        self._set_account("AllAccounts", "")

    def _cleanup(self) -> None:
        self._shortcut.set_action(None)
        self.unregister_events()
        get_log_console_handler().set_callback(None)

    def _on_adj_upper_changed(
        self, adj: Gtk.Adjustment, _pspec: GObject.ParamSpec
    ) -> None:
        if adj.get_upper() == adj.get_page_size():
            self._ui.jump_to_end_button.set_visible(False)

    def _on_adj_value_changed(
        self, adj: Gtk.Adjustment, _pspec: GObject.ParamSpec
    ) -> None:
        bottom = adj.get_upper() - adj.get_page_size()
        autoscroll = bottom - adj.get_value() < 1
        self._ui.jump_to_end_button.set_visible(not autoscroll)

    def _on_jump_to_end_clicked(self, _button: Gtk.Button) -> None:
        vadjustment = self._ui.scrolled.get_vadjustment()
        vadjustment.set_value(vadjustment.get_upper())

    def _on_account_change(self, dropdown: GajimDropDown[str], *args: Any) -> None:
        item = dropdown.get_selected_item()
        assert item is not None
        self._selected_send_account = item.key

    def _set_title(self) -> None:
        if self._selected_account == "AllAccounts":
            title = _("All Accounts")
        elif self._selected_account == "AccountWizard":
            title = _("Account Wizard")
        else:
            title = app.get_jid_from_account(self._selected_account)
        self.set_title(title)

    def _on_account_changed(self, event: AccountEnabled | AccountDisabled) -> None:
        buf = self._ui.protocol_view.get_buffer()

        if isinstance(event, AccountEnabled):
            buf.create_tag(event.account)
        else:
            start, end = buf.get_bounds()
            buf.remove_tag_by_name(event.account, start, end)

    def _on_stack_child_changed(
        self, _widget: Gtk.Stack, _pspec: GObject.ParamSpec
    ) -> None:
        name = self._ui.stack.get_visible_child_name()
        self._ui.search_toggle.set_sensitive(name == "protocol")

    def _create_tags(self) -> None:
        tags = ["incoming", "outgoing", "presence", "message", "stream", "iq"]

        accounts = app.settings.get_active_accounts()
        tags.extend(accounts)

        tags.append("AccountWizard")

        for tag_name in tags:
            self._ui.protocol_view.get_buffer().create_tag(tag_name)

    def _add_stanza_presets(self) -> None:
        for stanza_type in STANZA_PRESETS:
            row = Gtk.ListBoxRow()
            label = Gtk.Label(label=stanza_type, halign=Gtk.Align.START)
            row.set_child(label)
            self._ui.stanza_presets_listbox.append(row)

        self._ui.stanza_presets_listbox.set_visible(True)

    def _add_log_record(self, message: str) -> None:
        buf = self._ui.log_view.get_buffer()
        end_iter = buf.get_end_iter()
        buf.insert(end_iter, message)

    def _on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval == Gdk.KEY_Escape:
            if self._ui.search_toggle.get_active():
                self._ui.search_toggle.set_active(False)
                return Gdk.EVENT_STOP

            self.close()

        if (
            state & Gdk.ModifierType.CONTROL_MASK and keyval == Gdk.KEY_Return
        ) or keyval == Gdk.KEY_KP_Enter:
            self._on_send()
            return Gdk.EVENT_STOP

        if state & Gdk.ModifierType.CONTROL_MASK and keyval == Gdk.KEY_Up:
            self._on_paste_previous()
            return Gdk.EVENT_STOP

        if state & Gdk.ModifierType.CONTROL_MASK and keyval == Gdk.KEY_Down:
            self._on_paste_next()
            return Gdk.EVENT_STOP

        if state & Gdk.ModifierType.CONTROL_MASK and keyval == Gdk.KEY_f:
            self._ui.search_toggle.set_active(True)
            self._ui.search_entry.grab_focus()
            return Gdk.EVENT_STOP

        if keyval == Gdk.KEY_F3:
            self._find(Direction.NEXT)
            return Gdk.EVENT_STOP

        if state & Gdk.ModifierType.SHIFT_MASK and keyval == Gdk.KEY_F3:
            self._find(Direction.PREV)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _on_row_activated(self, _listbox: Gtk.ListBox, row: Gtk.ListBoxRow) -> None:
        child = row.get_child()
        assert isinstance(child, Gtk.Label)
        text = child.get_text()

        stanza_string = STANZA_PRESETS.get(text)

        if stanza_string is not None:
            buffer_ = self._ui.input_entry.get_buffer()
            buffer_.set_text(stanza_string)
            self._ui.input_entry.grab_focus()

    def _on_send(self, *args: Any) -> None:
        if not self._selected_send_account:
            return
        if not app.account_is_available(self._selected_send_account):
            # If offline or connecting
            InformationAlertDialog(
                _("You are offline"),
                _('Please make sure you are connected with "%s".')
                % self._selected_send_account,
            )
            return
        buffer_ = self._ui.input_entry.get_buffer()
        begin_iter, end_iter = buffer_.get_bounds()
        stanza = buffer_.get_text(begin_iter, end_iter, True)
        if stanza:
            try:
                node = nbxmpp.Node(node=stanza)
            except Exception as error:
                InformationAlertDialog(_("Invalid Node"), str(error))
                return

            if node.getName() in ("message", "presence", "iq"):
                # Parse stanza again if its a message, presence or iq and
                # set jabber:client as toplevel namespace
                # Use type Protocol so nbxmpp counts the stanza for
                # stream management
                try:
                    node = nbxmpp.Protocol(
                        node=stanza, attrs={"xmlns": "jabber:client"}
                    )
                except Exception as error:
                    InformationAlertDialog(_("Invalid Stanza"), str(error))
                    return

            client = app.get_client(self._selected_send_account)
            assert isinstance(node, nbxmpp.Protocol)
            client.connection.send_stanza(node)
            self._sent_stanzas.add(stanza)
            buffer_.set_text("")

    def _on_paste_previous(self, *args: Any) -> None:
        buffer_ = self._ui.input_entry.get_buffer()
        buffer_.set_text(self._sent_stanzas.get_previous())
        self._ui.input_entry.grab_focus()

    def _on_paste_next(self, *args: Any) -> None:
        buffer_ = self._ui.input_entry.get_buffer()
        buffer_.set_text(self._sent_stanzas.get_next())
        self._ui.input_entry.grab_focus()

    def _on_input(self, button: Gtk.ToggleButton) -> None:
        child2 = self._ui.paned.get_end_child()
        assert child2 is not None
        active = button.get_active()

        child2.set_visible(active)
        self._ui.send.set_visible(active)
        self._ui.paste.set_visible(active)
        self._ui.account_label.set_visible(active)
        self._account_dropdown.set_visible(active)
        self._ui.menubutton.set_visible(active)

        if active:
            self._ui.input_entry.grab_focus()

    def _on_search_toggled(self, button: Gtk.ToggleButton) -> None:
        self._ui.search_revealer.set_reveal_child(button.get_active())
        self._ui.search_entry.grab_focus()

    def _on_search_activate(self, _entry: Gtk.SearchEntry) -> None:
        self._find(Direction.NEXT)

    def _on_search_clicked(self, button: Any) -> None:
        if button is self._ui.search_forward:
            direction = Direction.NEXT
        else:
            direction = Direction.PREV
        self._find(direction)

    def _find(self, direction: Direction) -> None:
        self._search_settings.set_search_text(self._ui.search_entry.get_text())
        textbuffer = self._ui.protocol_view.get_buffer()

        last_pos_mark = textbuffer.get_mark("last_pos")
        if last_pos_mark is not None:
            current_pos = textbuffer.get_iter_at_mark(last_pos_mark)
        else:
            current_pos = textbuffer.get_start_iter()

        if direction == Direction.NEXT:
            self._search_context.forward_async(
                current_pos, None, self._on_search_finished, direction
            )
        else:
            self._search_context.backward_async(
                current_pos, None, self._on_search_finished, direction
            )

    def _on_search_finished(
        self,
        _context: GtkSource.SearchContext,
        result: Gio.AsyncResult,
        direction: Direction,
    ) -> None:
        if direction == Direction.NEXT:
            match = self._search_context.forward_finish(result)
        else:
            match = self._search_context.backward_finish(result)

        match_found, match_start, match_end, _has_wrapped_around = match

        if not match_found:
            self._ui.search_results_label.set_text(_("No results"))
            return

        occurrences_count = self._search_context.get_occurrences_count()
        if occurrences_count == -1:
            # Text scan may not be complete yet
            occurrences_count = "?"

        occurrence_positon = self._search_context.get_occurrence_position(
            match_start, match_end
        )
        if occurrence_positon == -1:
            occurrence_positon = 1
        self._ui.search_results_label.set_text(
            _("%s of %s") % (occurrence_positon, occurrences_count)
        )

        textbuffer = self._ui.protocol_view.get_buffer()

        if direction == Direction.NEXT:
            mark = textbuffer.create_mark("last_pos", match_end, True)
        else:
            mark = textbuffer.create_mark("last_pos", match_start, True)
        self._ui.protocol_view.scroll_to_mark(mark, 0, True, 0.5, 0.5)

    @staticmethod
    def _get_accounts() -> dict[str, str]:
        """Gets a dictionary of accounts ready to use with GajimDropdown.
        All available accounts are added, connected accounts are placed first.
        Additionally, AccountWizard is added as a special account.
        """
        dropdown_data: dict[str, str] = {}

        connected_accounts = app.get_connected_accounts()
        for account in connected_accounts:
            label = app.get_account_label(account)
            dropdown_data[account] = label

        for account in app.get_accounts_sorted():
            if account in connected_accounts:
                continue
            label = app.get_account_label(account)
            dropdown_data[account] = label

        dropdown_data["AccountWizard"] = _("Account Wizard")

        return dropdown_data

    def _on_filter_options(self, _button: Gtk.Button) -> None:
        if self._filter_dialog is not None:
            self._filter_dialog.present()
            return

        all_accounts = {"AllAccounts": _("All Accounts"), **self._get_accounts()}

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Account"),
                SettingType.VALUE,
                self._selected_account,
                callback=self._set_account,
                props={"data": all_accounts},
            ),
            Setting(
                SettingKind.SWITCH,
                "Presence",
                SettingType.VALUE,
                self._presence,
                callback=self._on_setting,
                data="presence",
            ),
            Setting(
                SettingKind.SWITCH,
                "Message",
                SettingType.VALUE,
                self._message,
                callback=self._on_setting,
                data="message",
            ),
            Setting(
                SettingKind.SWITCH,
                "IQ",
                SettingType.VALUE,
                self._iq,
                callback=self._on_setting,
                data="iq",
            ),
            Setting(
                SettingKind.SWITCH,
                "Stream Management",
                SettingType.VALUE,
                self._stream,
                callback=self._on_setting,
                data="stream",
            ),
            Setting(
                SettingKind.SWITCH,
                "In",
                SettingType.VALUE,
                self._incoming,
                callback=self._on_setting,
                data="incoming",
            ),
            Setting(
                SettingKind.SWITCH,
                "Out",
                SettingType.VALUE,
                self._outgoing,
                callback=self._on_setting,
                data="outgoing",
            ),
        ]

        self._filter_dialog = SettingsDialog(
            self,
            _("Filter"),
            Gtk.DialogFlags.DESTROY_WITH_PARENT,
            settings,
            self._selected_account or "AllAccounts",
        )
        self._filter_dialog.connect("close-request", self._on_filter_destroyed)

    def _on_filter_destroyed(self, window: Gtk.Window) -> None:
        window.disconnect_by_func(self._on_filter_destroyed)
        self._filter_dialog = None

    def _on_clear_window(self, *args: Any) -> None:
        self._ui.protocol_view.get_buffer().set_text("")

    def _apply_filters(self) -> None:
        table = self._ui.protocol_view.get_buffer().get_tag_table()

        active_accounts = app.settings.get_active_accounts()
        active_accounts.append("AccountWizard")
        for account in active_accounts:
            tag = table.lookup(account)
            if tag is None:
                continue

            if self._selected_account == "AllAccounts":
                visible = True
            else:
                visible = account == self._selected_account

            if visible:
                tag.set_priority(0)
            else:
                tag.set_priority(table.get_size() - 1)

            tag.set_property("invisible", not visible)

        for data in ["presence", "message", "iq", "stream", "incoming", "outgoing"]:
            visible = getattr(self, f"_{data}")

            tag = table.lookup(data)
            if tag is None:
                continue

            if data in ("incoming", "outgoing"):
                if visible:
                    tag.set_priority(0)
                else:
                    tag.set_priority(table.get_size() - 1)

            tag.set_property("invisible", not visible)

    def _set_account(self, value: str, _data: Any) -> None:
        self._selected_account = value
        self._set_title()
        self._apply_filters()

    def _on_setting(self, value: bool, data: str) -> None:
        setattr(self, f"_{data}", value)
        self._apply_filters()

    def _on_stanza_received(self, event: StanzaReceived):
        self._print_stanza(event, "incoming")

    def _on_stanza_sent(self, event: StanzaSent):
        self._print_stanza(event, "outgoing")

    def _print_stanza(self, event: StanzaReceived | StanzaSent, kind: str) -> None:
        if event.account == "AccountWizard":
            account_label = _("Account Wizard")
        else:
            account_label = app.get_account_label(event.account)

        stanza = event.stanza
        if not isinstance(stanza, str):
            # pylint: disable=unnecessary-dunder-call
            stanza = stanza.__str__(fancy=True)

        if not stanza:
            return

        is_at_the_end = at_the_end(self._ui.scrolled)

        buffer_ = self._ui.protocol_view.get_buffer()
        end_iter = buffer_.get_end_iter()

        type_ = kind
        if stanza.startswith("<presence"):
            type_ = "presence"
        elif stanza.startswith("<message"):
            type_ = "message"
        elif stanza.startswith("<iq"):
            type_ = "iq"
        elif stanza.startswith(("<r", "<a")):
            type_ = "stream"

        text = "<!-- {kind} {time} ({account}) -->\n{stanza}\n\n".format(
            kind=kind.capitalize(),
            time=time.strftime("%c"),
            account=account_label,
            stanza=stanza,
        )
        buffer_.insert_with_tags_by_name(end_iter, text, type_, kind, event.account)

        if is_at_the_end:
            GLib.idle_add(scroll_to, self._ui.scrolled, "bottom")


class SentSzanzas:
    def __init__(self) -> None:
        self._sent_stanzas: dict[float, str] = {}
        self._last_selected_ts = 0

    def add(self, stanza: str) -> None:
        self._sent_stanzas[time.time()] = stanza
        self._last_selected_ts = 0

    def get_previous(self) -> str:
        return self._get(Direction.PREV)

    def get_next(self) -> str:
        return self._get(Direction.NEXT)

    def _get(self, direction: Direction) -> str:
        if not self._sent_stanzas:
            return ""

        if direction == Direction.PREV:
            for timestamp, stanza in reversed(self._sent_stanzas.items()):
                if timestamp >= self._last_selected_ts:
                    continue
                self._last_selected_ts = timestamp
                return stanza
        else:
            for timestamp, stanza in self._sent_stanzas.items():
                if timestamp <= self._last_selected_ts:
                    continue
                self._last_selected_ts = timestamp
                return stanza

        self._last_selected_ts = list(self._sent_stanzas.keys())[-1]
        return self._sent_stanzas[self._last_selected_ts]



# Copyright (C) 2009 Bruno Tarquini <btarquini AT gmail.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import logging
import os
import sys
import time
from collections.abc import Callable
from datetime import datetime

from gajim.common import app
from gajim.common import configpaths
from gajim.common.i18n import _

LogCallback = Callable[[str], None]


def parseLogLevel(arg: str) -> int:
    '''
    Either numeric value or level name from logging module
    '''
    if arg.isdigit():
        return int(arg)
    if arg.isupper() and hasattr(logging, arg):
        return getattr(logging, arg)
    print(_('%s is not a valid loglevel') % repr(arg), file=sys.stderr)
    return 0


def parseLogTarget(arg: str) -> str:
    '''
    [gajim.]c.x.y  ->  gajim.c.x.y
    .other_logger  ->  other_logger
    <None>         ->  gajim
    '''
    arg = arg.lower()
    if not arg:
        return 'gajim'
    if arg.startswith('.'):
        return arg[1:]
    if arg.startswith('gajim'):
        return arg
    return 'gajim.' + arg


def parseAndSetLogLevels(arg: str) -> None:
    '''
    [=]LOGLEVEL     ->  gajim=LOGLEVEL
    gajim=LOGLEVEL  ->  gajim=LOGLEVEL
    .other=10       ->  other=10
    .=10            ->  <nothing>
    c.x.y=c.z=20    ->  gajim.c.x.y=20
                        gajim.c.z=20
    gajim=10,c.x=20 ->  gajim=10
                        gajim.c.x=20
    '''
    for directive in arg.split(','):
        directive = directive.strip()
        if not directive:
            continue
        if '=' not in directive:
            directive = '=' + directive
        targets, level = directive.rsplit('=', 1)
        level = parseLogLevel(level.strip())
        for target in targets.split('='):
            target = parseLogTarget(target.strip())
            if target:
                logging.getLogger(target).setLevel(level)
                print('Logger %s level set to %d' % (target, level),
                      file=sys.stderr)


class Colors:
    NONE = chr(27) + '[0m'
    BLACK = chr(27) + '[30m'
    RED = chr(27) + '[31m'
    GREEN = chr(27) + '[32m'
    BROWN = chr(27) + '[33m'
    BLUE = chr(27) + '[34m'
    MAGENTA = chr(27) + '[35m'
    CYAN = chr(27) + '[36m'
    LIGHT_GRAY = chr(27) + '[37m'
    DARK_GRAY = chr(27) + '[30;1m'
    BRIGHT_RED = chr(27) + '[31;1m'
    BRIGHT_GREEN = chr(27) + '[32;1m'
    YELLOW = chr(27) + '[33;1m'
    BRIGHT_BLUE = chr(27) + '[34;1m'
    PURPLE = chr(27) + '[35;1m'
    BRIGHT_CYAN = chr(27) + '[36;1m'
    WHITE = chr(27) + '[37;1m'


def colorize(text: str, color: str) -> str:
    return color + text + Colors.NONE


class LogConsoleHandler(logging.StreamHandler):  # pyright: ignore
    def __init__(self) -> None:
        super().__init__()  # pyright: ignore
        self._callback: LogCallback | None = None

    def emit(self, record: logging.LogRecord) -> None:
        if record.levelno < logging.WARNING:
            return

        msg = self.format(record)
        app.logging_records.append(msg)
        if self._callback is not None:
            self._callback(msg)

    def set_callback(self,
                     func: LogCallback | None
                     ) -> None:
        self._callback = func


logging.Formatter.default_msec_format = None

class FancyFormatter(logging.Formatter):
    '''
    An eye-candy formatter with Colors
    '''
    colors_mapping = {
        'DEBUG': Colors.BLUE,
        'INFO': Colors.GREEN,
        'WARNING': Colors.BROWN,
        'ERROR': Colors.RED,
        'CRITICAL': Colors.BRIGHT_RED,
    }

    def __init__(self,
                 fmt: str | None = None,
                 use_color: bool = False) -> None:
        logging.Formatter.__init__(self, fmt)
        self.use_color = use_color

    def format(self, record: logging.LogRecord) -> str:
        level = record.levelname
        record.levelname = '(%s)' % level[0]

        if self.use_color:
            c = FancyFormatter.colors_mapping.get(level, '')
            record.levelname = colorize(record.levelname, c)
            record.name = '%-25s' % colorize(record.name, Colors.CYAN)
        else:
            record.name = '%-25s|' % record.name

        return logging.Formatter.format(self, record)


def init() -> None:
    '''
    Initialize the logging system
    '''

    if app.get_debug_mode():
        _cleanup_debug_logs()
        _redirect_output()

    use_color = False
    if os.name != 'nt':
        use_color = sys.stderr.isatty()

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(
        FancyFormatter(
            '%(asctime)s %(levelname)s %(name)-35s %(message)s',
            use_color
        )
    )

    root_log = logging.getLogger('gajim')
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    root_log = logging.getLogger('nbxmpp')
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    root_log = logging.getLogger('gnupg')
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    root_log = logging.getLogger('omemo_dr')
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    if os.environ.get('GAJIM_LEAK'):
        logging.getLogger('gajim.leak').setLevel(logging.DEBUG)

    # GAJIM_DEBUG is set only on Windows when using Gajim-Debug.exe
    # Gajim-Debug.exe shows a command line prompt and we want to redirect
    # log output to it
    if app.get_debug_mode() or os.environ.get('GAJIM_DEBUG', False):  # noqa: PLW1508
        set_verbose()


def set_loglevels(loglevels_string: str) -> None:
    parseAndSetLogLevels(loglevels_string)


def set_verbose() -> None:
    parseAndSetLogLevels('gajim=DEBUG')
    parseAndSetLogLevels('.nbxmpp=INFO')
    parseAndSetLogLevels('.omemo_dr=DEBUG')


def set_quiet() -> None:
    parseAndSetLogLevels('gajim=CRITICAL')
    parseAndSetLogLevels('.nbxmpp=CRITICAL')
    parseAndSetLogLevels('.omemo_dr=CRITICAL')


def get_log_console_handler() -> LogConsoleHandler:
    return _log_console_handler


def _redirect_output() -> None:
    debug_folder = configpaths.get('DEBUG')
    date = datetime.today().strftime('%Y-%m-%d-%H%M%S')
    filename = '%s-debug.log' % date
    fd = open(debug_folder / filename, 'a', encoding='utf8')
    sys.stderr = sys.stdout = fd


def _cleanup_debug_logs() -> None:
    debug_folder = configpaths.get('DEBUG')
    debug_files = list(debug_folder.glob('*-debug.log*'))
    now = time.time()
    for file in debug_files:
        # Delete everything older than 3 days
        if file.stat().st_ctime < now - 259200:
            file.unlink()


_log_console_handler = LogConsoleHandler()
_log_console_handler.setFormatter(
    logging.Formatter(
        '%(asctime)s (%(levelname).1s) %(name)-35s | %(message)s\n',
    )
)


if __name__ == '__main__':
    init()

    set_loglevels('gajim.c=DEBUG,INFO')

    log = logging.getLogger('gajim')
    log.debug('debug')
    log.info('info')
    log.warning('warn')
    log.error('error')
    log.critical('critical')

    log = logging.getLogger('gajim.c.x.dispatcher')
    log.debug('debug')
    log.info('info')
    log.warning('warn')
    log.error('error')
    log.critical('critical')



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import time

from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.const import FTState
from gajim.common.i18n import _
from gajim.common.modules.httpupload import HTTPFileTransfer
from gajim.common.util.datetime import utc_now
from gajim.common.util.text import format_eta

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel


class FileTransferRow(BaseRow):
    def __init__(self, account: str, transfer: HTTPFileTransfer) -> None:
        BaseRow.__init__(self, account)

        self.type = "file-transfer"
        timestamp = utc_now()
        self.timestamp = timestamp.astimezone()
        self.db_timestamp = timestamp.timestamp()

        self._destroyed: bool = False

        if app.settings.get("use_kib_mib"):
            self._units = GLib.FormatSizeFlags.IEC_UNITS
        else:
            self._units = GLib.FormatSizeFlags.DEFAULT

        self._start_time = time.time()
        self._pulse = GLib.timeout_add(100, self._pulse_progressbar)

        self._transfer = transfer
        self._transfer.connect("state-changed", self._on_transfer_state_change)
        self._transfer.connect("progress", self._on_transfer_progress)

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 1)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_halign(Gtk.Align.START)
        timestamp_widget.set_valign(Gtk.Align.START)
        self.grid.attach(timestamp_widget, 1, 0, 1, 1)

        self._ui = get_builder("file_transfer.ui")
        self.grid.attach(self._ui.transfer_box, 1, 1, 1, 1)

        self._connect(self._ui.cancel_button, "clicked", self._on_cancel_clicked)

        self._ui.file_name.set_text(transfer.filename)
        self._ui.transfer_description.set_text(transfer.get_state_description())

    def do_unroot(self) -> None:
        self._destroyed = True

        self._transfer.disconnect_all_from_obj(self)
        del self._transfer
        if self._pulse is not None:
            GLib.source_remove(self._pulse)

        BaseRow.do_unroot(self)

    def _on_cancel_clicked(self, _button: Gtk.Button) -> None:
        if self._transfer.state.is_active:
            self._transfer.cancel()

        self.emit("remove")

    def _on_transfer_state_change(
        self, transfer: HTTPFileTransfer, _signal_name: str, state: FTState
    ) -> None:
        if self._destroyed:
            return

        if state.is_error:
            InformationAlertDialog(_("Error"), transfer.error_text)
            self.emit("remove")
            return

        if state.is_finished or state.is_cancelled:
            self.emit("remove")
            return

        description = transfer.get_state_description()
        if description:
            self._ui.transfer_description.set_text(description)

    def _pulse_progressbar(self):
        self._ui.progress_bar.pulse()
        return True

    def _on_transfer_progress(
        self, transfer: HTTPFileTransfer, _signal_name: str
    ) -> None:
        if self._destroyed:
            return
        if self._pulse is not None:
            GLib.source_remove(self._pulse)
            self._pulse = None

        time_now = time.time()

        size_total = GLib.format_size_full(transfer.size, self._units)
        self._ui.file_size.set_text(size_total)

        progress = transfer.get_progress()
        fraction = progress / transfer.size

        bytes_sec = int(round(progress / (time_now - self._start_time), 1))
        speed = f"{GLib.format_size_full(bytes_sec, self._units)}/s"
        self._ui.transfer_progress.set_tooltip_text(_("Speed: %s") % speed)

        if bytes_sec == 0:
            eta = "∞"
        else:
            eta = format_eta(round((transfer.size - progress) / bytes_sec))

        self._ui.transfer_progress.set_text(
            _("%(progress)s %% (%(time)s remaining)")
            % {"progress": round(fraction * 100), "time": eta}
        )

        self._ui.progress_bar.set_fraction(fraction)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import AvatarSize
from gajim.common.const import SimpleClientState
from gajim.common.ged import EventHelper
from gajim.common.util.status import get_client_status
from gajim.common.util.status import get_uf_show

from gajim.gtk.avatar import get_show_circle
from gajim.gtk.util.misc import convert_surface_to_texture
from gajim.gtk.util.misc import get_ui_string


@Gtk.Template(string=get_ui_string("status_selector.ui"))
class StatusSelector(Gtk.MenuButton, EventHelper):
    __gtype_name__ = "StatusSelector"

    _image: Gtk.Image = Gtk.Template.Child()
    _label: Gtk.Label = Gtk.Template.Child()

    def __init__(self):
        self._account: str | None = None

        Gtk.MenuButton.__init__(self)
        EventHelper.__init__(self)

        surface = get_show_circle(
            "offline", AvatarSize.SHOW_CIRCLE, self.get_scale_factor()
        )
        self._image.set_from_paintable(convert_surface_to_texture(surface))

        self.register_events(
            [
                ("our-show", ged.GUI1, self._on_our_show),
            ]
        )

    @GObject.Property(type=str)
    def account(self) -> str | None:  # pyright: ignore
        return self._account

    @account.setter
    def account(self, account: str | None) -> None:
        if self._account == account:
            return

        self._disconnect_signals(self._account)
        self._connect_signals(account)

        self._account = account
        self._update_status()

    def _connect_signals(self, account: str | None) -> None:
        if account is None:
            return
        client = app.get_client(account)
        client.connect_signal("state-changed", self._on_client_state_changed)

    def _disconnect_signals(self, account: str | None) -> None:
        if account is None:
            return
        client = app.get_client(account)
        client.disconnect_all_from_obj(self)

    def set_account(self, account: str | None) -> None:
        self.account = account

    @Gtk.Template.Callback()
    def _on_clicked(self, _button: Gtk.Button, status: str) -> None:
        app.app.change_status(status=status, account=self._account)

    def _on_our_show(self, event: events.ShowChanged) -> None:
        if event.account != self._account:
            return
        self._update_status()

    def _on_client_state_changed(
        self, _client: Client, _signal_name: str, _state: SimpleClientState
    ) -> None:
        self._update_status()

    def _update_status(self) -> None:
        if self._account is None:
            self._image.set_from_paintable(None)
            self._label.set_text("")
            return

        show = get_client_status(self._account)
        surface = get_show_circle(show, AvatarSize.SHOW_CIRCLE, self.get_scale_factor())
        self._image.set_from_paintable(convert_surface_to_texture(surface))
        self._label.set_text(get_uf_show(show))


@Gtk.Template(string=get_ui_string("status_selector_popover.ui"))
class StatusSelectorPopover(Gtk.Popover):
    __gtype_name__ = "StatusSelectorPopover"

    __gsignals__ = {
        "clicked": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (str,),
        )
    }

    _box: Gtk.Box = Gtk.Template.Child()

    @Gtk.Template.Callback()
    def _on_clicked(self, button: StatusSelectorPopoverButton) -> None:
        self.emit("clicked", button.status)
        self.popdown()


@Gtk.Template(string=get_ui_string("status_selector_popover_button.ui"))
class StatusSelectorPopoverButton(Gtk.Button):
    __gtype_name__ = "StatusSelectorPopoverButton"

    _image: Gtk.Image = Gtk.Template.Child()
    _label: Gtk.Label = Gtk.Template.Child()

    @GObject.Property(type=str)
    def status(self) -> str:  # pyright: ignore
        return self._status

    @status.setter
    def status(self, status: str) -> None:
        self._status = status
        self._label.set_text_with_mnemonic(get_uf_show(status, use_mnemonic=True))
        surface = get_show_circle(
            status, AvatarSize.SHOW_CIRCLE, self.get_scale_factor()
        )
        self._image.set_from_paintable(convert_surface_to_texture(surface))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.alert import AlertDialog
from gajim.gtk.alert import DialogResponse
from gajim.gtk.settings import GajimPreferencePage
from gajim.gtk.settings import GajimPreferencesGroup
from gajim.gtk.shortcut_manager import GajimShortcut
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

# All keyvals:
# https://gitlab.gnome.org/GNOME/gtk/-/blob/main/gdk/gdkkeysyms.h

SHORTCUT_CATEGORIES = {
    "general": _("General"),
    "chats": _("Chats"),
    "messages": _("Messages"),
}


class ShortcutsPage(GajimPreferencePage, SignalManager):
    key = "shortcuts"
    icon_name = "lucide-keyboard-symbolic"
    label = _("Shortcuts")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Shortcuts"),
            groups=[],
        )
        SignalManager.__init__(self)

        preferences_groups: dict[str, GajimPreferencesGroup] = {}
        for category, title in SHORTCUT_CATEGORIES.items():
            preferences_group = GajimPreferencesGroup(key=category, title=title)
            preferences_groups[category] = preferences_group
            self.add(preferences_group)

        self._manager = app.app.get_shortcut_manager()
        for shortcut in self._manager.iter_shortcuts():
            if not shortcut.allow_rebind:
                continue

            row = ShortcutRow(shortcut)
            self._connect(row, "activated", self._on_activated)

            preferences_groups[shortcut.category].add(row)

    def do_unroot(self) -> None:
        self._disconnect_all()
        GajimPreferencePage.do_unroot(self)

    def _on_activated(self, row: ShortcutRow) -> None:
        parent = cast(Adw.ApplicationWindow, self.get_root())
        dialog = KeyEntryDialog(parent=parent)
        dialog.connect("response", self._on_shortcut_edited, row.shortcut)

    def _on_shortcut_edited(
        self,
        dialog: KeyEntryDialog,
        response_id: str,
        shortcut: GajimShortcut,
    ) -> None:
        if response_id == "close":
            return

        if response_id == "reset":
            shortcut.reset()
            self._manager.store_user_shortcuts()
            return

        shortcut.set_custom_accelerators(dialog.get_accelerators())
        self._manager.store_user_shortcuts()


@Gtk.Template(string=get_ui_string("shortcut_row.ui"))
class ShortcutRow(Adw.ActionRow, SignalManager):
    __gtype_name__ = "ShortcutRow"

    _accelerator_badge: Gtk.Label = Gtk.Template.Child()
    _accelerator_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, shortcut: GajimShortcut) -> None:
        Adw.ActionRow.__init__(self)
        SignalManager.__init__(self)

        self._shortcut = shortcut
        self.set_activatable(True)
        self.set_title(shortcut.label)
        self._update_fields()

        self._connect(self._shortcut, "notify::trigger", self._on_trigger_changed)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Adw.ActionRow.do_unroot(self)

    def _update_fields(self):
        self._accelerator_badge.set_visible(self._shortcut.has_custom_accel())
        self._accelerator_label.set_label(self._shortcut.get_accel_label())

    def _on_trigger_changed(
        self, shortcut: GajimShortcut, _pspec: GObject.ParamSpec
    ) -> None:
        self._update_fields()

    @property
    def shortcut(self) -> GajimShortcut:
        return self._shortcut


class KeyEntryDialog(AlertDialog):
    def __init__(self, parent: Gtk.Window) -> None:
        AlertDialog.__init__(
            self,
            heading=_("Edit Shortcut"),
            body=_(
                "Press the keys you want to associate with this action. "
                "To disable this shortcut, click apply."
            ),
            responses=[
                DialogResponse("close", _("Cancel")),
                DialogResponse("reset", _("Reset")),
                DialogResponse("apply", _("Apply"), "suggested", True),
            ],
            parent=parent,
        )

        self._accelerator_name = ""

        controller = Gtk.EventControllerKey(
            propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        controller.connect("key-pressed", self._on_key_pressed)
        self.add_controller(controller)

        self._shortcut_label = Gtk.Label(label=_("No keys pressed"))
        self._shortcut_label.add_css_class("bold")
        self._shortcut_label.add_css_class("accent")
        self._shortcut_label.add_css_class("py-6")
        self.set_extra_child(self._shortcut_label)

    def get_accelerators(self) -> list[str] | None:
        if self._accelerator_name:
            return [self._accelerator_name]
        return None

    def _on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if not Gtk.accelerator_valid(keyval, state):
            # Modifier combination (e.g. Alt_L + Control_L)
            return Gdk.EVENT_STOP

        self._accelerator_name = Gtk.accelerator_name(keyval, state)
        self._shortcut_label.set_label(Gtk.accelerator_get_label(keyval, state))
        return Gdk.EVENT_STOP



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import GLib
from gi.repository import Gtk

from gajim.common.const import AvatarSize
from gajim.common.util.datetime import utc_now
from gajim.common.util.text import process_non_spacing_marks

from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import SimpleLabel


class CommandOutputRow(BaseRow):
    def __init__(self, account: str, text: str, is_error: bool) -> None:
        BaseRow.__init__(self, account)

        self.type = "command_output"
        now = utc_now()
        self.timestamp = now.astimezone()
        self.db_timestamp = now.timestamp()

        self.add_css_class("conversation-command-row")

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        icon = Gtk.Image.new_from_icon_name("lucide-terminal-symbolic")
        icon.set_pixel_size(32)
        icon.add_css_class("dimmed")
        avatar_placeholder.append(icon)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 1)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_valign(Gtk.Align.START)
        timestamp_widget.set_margin_start(0)
        self.grid.attach(timestamp_widget, 1, 0, 1, 1)

        text = GLib.markup_escape_text(text)
        markup = f"<tt>{text}</tt>"
        self._label = SimpleLabel()
        if is_error:
            self._label.add_css_class("gajim-command-error")
        else:
            self._label.add_css_class("gajim-command-output")
        self._label.set_markup(process_non_spacing_marks(markup))
        self.grid.attach(self._label, 1, 1, 1, 1)

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging
from datetime import timedelta

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.dataforms import MultipleDataForm
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import LastActivityData
from nbxmpp.structs import SoftwareVersionResult
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.const import SettingType
from gajim.gtk.preference.widgets import CopyButton
from gajim.gtk.preference.widgets import PlaceholderBox
from gajim.gtk.settings import GajimPreferencePage
from gajim.gtk.settings import GajimPreferencesGroup
from gajim.gtk.settings import SubPageSetting
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.misc import open_uri

log = logging.getLogger("gajim.gtk.preference.server_info")


@Gtk.Template.from_string(string=get_ui_string("preference/provider_info.ui"))
class AccountProviderInfoGroup(Adw.PreferencesGroup, SignalManager):
    __gtype_name__ = "AccountProviderInfoGroup"

    _clipboard_button: CopyButton = Gtk.Template.Child()
    _hostname_row: Adw.ActionRow = Gtk.Template.Child()
    _software_row: Adw.ActionRow = Gtk.Template.Child()
    _uptime_row: Adw.ActionRow = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        Adw.PreferencesGroup.__init__(self)
        SignalManager.__init__(self)

        self._hostname = app.get_hostname_from_account(account, prefer_custom=True)
        self._software = ""
        self._uptime = ""

        self._hostname_row.set_subtitle(self._hostname)

        client = app.get_client(account)
        domain = client.get_own_jid().domain
        client.get_module("SoftwareVersion").request_software_version(
            domain, callback=self._software_version_received
        )

        client.get_module("LastActivity").request_last_activity(
            domain, callback=self._on_last_activity
        )

        self.add(
            SubPageSetting(
                account,
                None,
                _("Contacts"),
                SettingType.DIALOG,
                None,
                None,
                subpage=f"{account}-provider-contacts",
            )
        )

        self._connect(
            self._clipboard_button, "clicked", self._on_clipboard_button_clicked
        )

    def _on_clipboard_button_clicked(self, _widget: Gtk.Button) -> None:
        string = _(
            "Hostname: %(hostname)s\nSoftware: %(software)s\nUptime: %(uptime)s\n"
        )

        app.window.get_clipboard().set(
            string
            % {
                "hostname": self._hostname,
                "software": self._software,
                "uptime": self._uptime,
            }
        )

    def _on_last_activity(self, task: Task) -> None:
        try:
            result = cast(LastActivityData, task.finish())
        except (StanzaError, MalformedStanzaError) as error:
            log.warning(error)
            return

        delta = timedelta(seconds=result.seconds)
        hours = 0
        if result.seconds >= 3600:
            hours = delta.seconds // 3600
        self._uptime = _("%(days)s days, %(hours)s hours") % {
            "days": delta.days,
            "hours": hours,
        }
        self._uptime_row.set_subtitle(self._uptime)

    def _software_version_received(self, task: Task) -> None:
        try:
            result = cast(SoftwareVersionResult, task.finish())
        except StanzaError:
            self._software = _("Unknown")
        else:
            self._software = f"{result.name} {result.version}"

        self._software_row.set_subtitle(self._software)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Adw.PreferencesGroup.do_unroot(self)


class AccountProviderContactsGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key=f"{account}-provider-contacts",
            account=account,
            title=_("Contacts"),
        )

        client = app.get_client(account)
        server_info = client.get_module("Discovery").server_info
        if server_info is None:
            return

        self._add_contact_addresses(server_info.dataforms)

    def _add_contact_addresses(
        self, dataforms: list[SimpleDataForm | MultipleDataForm]
    ) -> None:
        fields = {
            "admin-addresses": _("Admin"),
            "status-addresses": _("Status"),
            "support-addresses": _("Support"),
            "security-addresses": _("Security"),
            "feedback-addresses": _("Feedback"),
            "abuse-addresses": _("Abuse"),
            "sales-addresses": _("Sales"),
        }

        addresses = self._get_addresses(fields, dataforms)
        if addresses is None:
            label = Gtk.Label(
                label=_("No contact addresses published for this server."),
                wrap=True,
                max_width_chars=50,
            )
            label.add_css_class("dimmed")
            label.add_css_class("p-18")
            self.add(label)
            return

        for address_type, values in addresses.items():
            contact_address_row = Adw.ActionRow(title=fields[address_type])

            addresses_box = Gtk.Box(
                orientation=Gtk.Orientation.VERTICAL, valign=Gtk.Align.CENTER, spacing=2
            )
            addresses_box.add_css_class("p-6")
            for value in values:
                label = self._get_address_label(value)
                addresses_box.append(label)

            contact_address_row.add_suffix(addresses_box)

            self.add(contact_address_row)

    @staticmethod
    def _get_addresses(
        fields: dict[str, str], dataforms: list[SimpleDataForm | MultipleDataForm]
    ) -> dict[str, list[str]] | None:
        addresses: dict[str, list[str]] = {}
        for form in dataforms:
            if not isinstance(form, SimpleDataForm):
                continue

            field = form.vars.get("FORM_TYPE")
            if field is None or field.value != "http://jabber.org/network/serverinfo":
                continue

            for address_type in fields:
                field = form.vars.get(address_type)
                if field is None:
                    continue

                if field.type_ != "list-multi":
                    continue

                if not field.values:
                    continue
                addresses[address_type] = cast(list[str], field.values)

            return addresses or None
        return None

    def _get_address_label(self, address: str) -> Gtk.Label:
        label = Gtk.Label(
            ellipsize=Pango.EllipsizeMode.END,
            halign=Gtk.Align.END,
            label=f'<a href="{address}">{address}</a>',
            use_markup=True,
            xalign=1,
        )

        label.add_css_class("link-button")
        self._connect(label, "activate-link", self._on_activate_link)
        return label

    def _on_activate_link(self, label: Gtk.Label, *args: Any) -> int:
        open_uri(label.get_text())
        return Gdk.EVENT_STOP


class AccountProviderFeaturesGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key=f"{account}-provider-features",
            account=account,
            title=_("Features"),
            description=_("Your service offers the following features"),
        )

        if app.settings.get("use_kib_mib"):
            self._units = GLib.FormatSizeFlags.IEC_UNITS
        else:
            self._units = GLib.FormatSizeFlags.DEFAULT

        self._add_features()
        self.add_copy_button()

    def do_unroot(self) -> None:
        GajimPreferencesGroup.do_unroot(self)
        self._features.clear()

    def _get_clipboard_text(self) -> str:
        string = ""
        for feature in self._features:
            string += feature.get_clipboard_text()
        return string

    def _add_features(self) -> None:
        assert self.account is not None
        client = app.get_client(self.account)

        http_upload_module = client.get_module("HTTPUpload")
        http_upload_info = http_upload_module.httpupload_namespace
        if http_upload_module.available:
            max_size = http_upload_module.max_file_size
            if max_size is not None:
                max_size = GLib.format_size_full(int(max_size), self._units)
                http_upload_info = f"{http_upload_info} (max. {max_size})"

        assert client.features is not None

        self._features = [
            FeatureRow("XEP-0045: Multi-User Chat", client.get_module("MUC").supported),
            FeatureRow(
                "XEP-0054: vcard-temp", client.get_module("VCardTemp").supported
            ),
            FeatureRow(
                "XEP-0077: In-Band Registration",
                client.get_module("Register").supported,
            ),
            FeatureRow(
                "XEP-0163: Personal Eventing Protocol",
                client.get_module("PEP").supported,
            ),
            FeatureRow(
                "XEP-0163: #publish-options",
                client.get_module("PubSub").publish_options,
            ),
            FeatureRow(
                "XEP-0191: Blocking Command",
                client.get_module("Blocking").supported,
                Namespace.BLOCKING,
            ),
            FeatureRow(
                "XEP-0198: Stream Management",
                client.features.has_sm(),
                Namespace.STREAM_MGMT,
            ),
            FeatureRow(
                "XEP-0258: Security Labels in XMPP",
                client.get_module("SecLabels").supported,
                Namespace.SECLABEL,
            ),
            FeatureRow(
                "XEP-0280: Message Carbons",
                client.get_module("Carbons").supported,
                Namespace.CARBONS,
            ),
            FeatureRow(
                "XEP-0313: Message Archive Management",
                client.get_module("MAM").available,
            ),
            FeatureRow(
                "XEP-0363: HTTP File Upload",
                client.get_module("HTTPUpload").available,
                http_upload_info,
            ),
            FeatureRow(
                "XEP-0398: Avatar Conversion",
                client.get_module("VCardAvatars").avatar_conversion_available,
            ),
            FeatureRow(
                "XEP-0411: Bookmarks Conversion",
                client.get_module("Bookmarks").conversion,
            ),
            FeatureRow(
                "XEP-0402: Bookmarks Compat",
                client.get_module("Bookmarks").compat,
            ),
            FeatureRow(
                "XEP-0402: Bookmarks Compat PEP",
                client.get_module("Bookmarks").compat_pep,
            ),
        ]

        for feature in self._features:
            self.add(feature)


class FeatureRow(Adw.ActionRow):
    def __init__(
        self, name: str, available: bool, additional: str | None = None
    ) -> None:
        Adw.ActionRow.__init__(self)

        self.set_title(name)
        self.set_subtitle(additional or "")

        self._available = available

        self._icon = Gtk.Image()
        self.add_prefix(self._icon)

        if available:
            self._icon.set_from_icon_name("lucide-check-symbolic")
            self._icon.add_css_class("success")
        else:
            self._icon.set_from_icon_name("lucide-x-symbolic")
            self._icon.add_css_class("error")

    def get_clipboard_text(self) -> str:
        if self._available:
            available = _("Available")
        else:
            available = _("Not available")

        string = f"{self.get_title()}: {available}"

        if additional := self.get_subtitle():
            string += f" ({additional})"

        return f"{string}\n"


class AccountProviderContactsPage(GajimPreferencePage):
    key = "provider-contacts"

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Provider Contacts - %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        if not app.account_is_available(account):
            self.set_content(PlaceholderBox(valign=Gtk.Align.CENTER))
            return

        self.add(AccountProviderContactsGroup(account))


class AccountProviderPage(GajimPreferencePage):
    key = "provider"
    icon_name = "lucide-server-symbolic"
    label = _("Service Provider")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Provider - %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        if not app.account_is_available(account):
            self.set_content(PlaceholderBox(valign=Gtk.Align.CENTER))
            return

        self.add(AccountProviderInfoGroup(account))
        self.add(AccountProviderFeaturesGroup(account))
